# Copyright (c) 2017 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from contextlib import contextmanager
import copy

from aim import aim_store
from aim.api import resource as aim_resource
from aim import context as aim_context
from neutron_lib import context
from neutron_lib.plugins import directory
from oslo_log import log

from gbpservice.neutron.db import api as db_api
from gbpservice.neutron.services.grouppolicy import (
    group_policy_driver_api as api)

LOG = log.getLogger(__name__)


class InternalValidationError(Exception):
    pass


class RollbackTransaction(Exception):
    pass


class ValidationManager(object):

    def __init__(self):
        # REVISIT: Defer until after validating config? Or pass in PD
        # & MD?
        self.core_plugin = directory.get_plugin()
        self.md = self.core_plugin.mechanism_manager.mech_drivers[
            'apic_aim'].obj
        self.pd = self.md.gbp_driver
        self.sfcd = None
        sfc_plugin = directory.get_plugin('sfc')
        if sfc_plugin:
            driver = sfc_plugin.driver_manager.drivers.get('aim')
            if driver:
                self.sfcd = driver.obj

    def validate(self, repair=False):
        self.output("Validating deployment, repair: %s" % repair)

        self.result = api.VALIDATION_PASSED
        self.repair = repair

        # REVISIT: Validate configuration.

        # Load project names from Keystone.
        self.md.project_details_cache.load_projects()

        # Start transaction.
        #
        # REVISIT: Set session's isolation level to serializable?
        self.actual_context = context.get_admin_context()
        try:
            with db_api.CONTEXT_WRITER.using(
                    self.actual_context) as session:
                self.actual_session = session
                self.aim_mgr = self.md.aim
                self.actual_aim_ctx = aim_context.AimContext(session)
                self.expected_session = ValidationSession(self)
                self.expected_aim_ctx = aim_context.AimContext(
                    None, ValidationAimStore(self))

                # Validate & repair GBP->Neutron mappings.
                if self.pd:
                    self.pd.validate_neutron_mapping(self)

                # Start with no expected or actual AIM resources or DB
                # records.
                self._expected_aim_resources = {}
                self._actual_aim_resources = {}
                self._expected_db_instances = {}
                self._db_instance_primary_keys = {}

                # Validate Neutron->AIM mapping, getting expected AIM
                # resources and DB records.
                self.md.validate_aim_mapping(self)

                # Validate GBP->AIM mapping, getting expected AIM
                # resources and DB records.
                if self.pd:
                    self.pd.validate_aim_mapping(self)

                # Validate SFC->AIM mapping, getting expected AIM
                # resources and DB records.
                if self.sfcd:
                    self.sfcd.validate_aim_mapping(self)

                # Validate that actual AIM resources match expected
                # AIM resources.
                self._validate_aim_resources()

                # Validate that actual DB instances match expected DB
                # instances.
                self._validate_db_instances()

                # Commit or rollback transaction.
                if self.result is api.VALIDATION_REPAIRED:
                    self.output("Committing repairs")
                else:
                    if (self.repair and
                        self.result is api.VALIDATION_FAILED_UNREPAIRABLE):
                        self.output("Rolling back attempted repairs")
                    raise RollbackTransaction()
        except RollbackTransaction:
            pass
        except Exception as exc:
            self.output("Validation failed with exception: %s - see log for "
                        "details" % exc)
            LOG.exception(exc)
            return api.VALIDATION_FAILED_WITH_EXCEPTION

        # Bind unbound ports outside transaction.
        if (self.repair and
            self.result is not api.VALIDATION_FAILED_UNREPAIRABLE):
            self.output("Binding unbound ports")
            self.md.bind_unbound_ports(self)

        self.output("Validation result: %s" % self.result)
        return self.result

    def output(self, msg):
        LOG.info(msg)
        print(msg)

    def register_aim_resource_class(self, resource_class):
        if resource_class not in self._expected_aim_resources:
            self._expected_aim_resources[resource_class] = {}
            self._actual_aim_resources[resource_class] = {
                tuple(resource.identity): resource
                for resource in self.aim_mgr.find(
                        self.actual_aim_ctx, resource_class)}

    def expect_aim_resource(self, resource, replace=False, remove=False):
        expected_resources = self._expected_aim_resources[resource.__class__]
        key = tuple(resource.identity)
        if remove:
            del expected_resources[key]
            return
        elif not replace and key in expected_resources:
            self.output("resource %s already expected" % resource)
            raise InternalValidationError()
        for attr_name, attr_type in resource.other_attributes.items():
            attr_type_type = attr_type['type']
            if attr_type_type == 'string':
                value = getattr(resource, attr_name)
                setattr(resource, attr_name,
                        str(value) if value else value)
            elif (attr_type_type == 'array' and
                attr_type['items']['type'] == 'string'):
                # REVISIT: May also need to dedup arrays of types
                # other than string.
                value = list(set(getattr(resource, attr_name)))
                setattr(resource, attr_name, value)
        expected_resources[key] = resource

    def expected_aim_resource(self, resource):
        expected_resources = self._expected_aim_resources[resource.__class__]
        key = tuple(resource.identity)
        return expected_resources.get(key)

    def expected_aim_resources(self, resource_class):
        return list(self._expected_aim_resources[resource_class].values())

    def actual_aim_resource(self, resource):
        actual_resources = self._actual_aim_resources[resource.__class__]
        key = tuple(resource.identity)
        return actual_resources.get(key)

    def actual_aim_resources(self, resource_class):
        return list(self._actual_aim_resources[resource_class].values())

    def register_db_instance_class(self, instance_class, primary_keys):
        self._expected_db_instances.setdefault(instance_class, {})
        self._db_instance_primary_keys[instance_class] = primary_keys

    def expect_db_instance(self, instance):
        instance_class = instance.__class__
        expected_instances = self._expected_db_instances[instance_class]
        primary_keys = self._db_instance_primary_keys[instance_class]
        key = tuple([getattr(instance, k) for k in primary_keys])
        expected_instances[key] = instance

    def query_db_instances(self, entities, args, filters):
        assert(1 == len(entities))
        assert(0 == len(args))
        instance_class = entities[0]
        expected_instances = self._expected_db_instances[instance_class]
        primary_keys = self._db_instance_primary_keys[instance_class]
        if filters:
            if (set(filters.keys()) == set(primary_keys)):
                key = tuple([filters[k] for k in primary_keys])
                instance = expected_instances.get(key)
                return [instance] if instance else []
            else:
                return [i for i in expected_instances.values()
                        if all([getattr(i, k) == v for k, v in
                                filters.items()])]
        else:
            return list(expected_instances.values())

    def should_repair(self, problem, action='Repairing'):
        if self.result is not api.VALIDATION_FAILED_UNREPAIRABLE:
            self.output("%s %s" % (action, problem))
            self.result = (api.VALIDATION_REPAIRED if self.repair
                           else api.VALIDATION_FAILED_REPAIRABLE)
            return True

    def validation_failed(self, reason):
        # REVISIT: Do we need drivers to be able to specify repairable
        # vs unrepairable? If so, add a keyword arg and make sure
        # VALIDATION_FAILED_REPAIRABLE does not overwrite
        # VALIDATION_FAILED_UNREPAIRABLE.
        self.output("Failed UNREPAIRABLE due to %s" % reason)
        self.result = api.VALIDATION_FAILED_UNREPAIRABLE

    def bind_ports_failed(self, message):
        self.output(message)
        self.result = api.VALIDATION_FAILED_BINDING_PORTS

    def _validate_aim_resources(self):
        for resource_class in self._expected_aim_resources.keys():
            self._validate_aim_resource_class(resource_class)

    def _validate_aim_resource_class(self, resource_class):
        expected_resources = self._expected_aim_resources[resource_class]

        for actual_resource in self.actual_aim_resources(resource_class):
            key = tuple(actual_resource.identity)
            expected_resource = expected_resources.pop(key, None)
            self._validate_actual_aim_resource(
                actual_resource, expected_resource)

        for expected_resource in expected_resources.values():
            self._handle_missing_aim_resource(expected_resource)

    def _validate_actual_aim_resource(self, actual_resource,
                                      expected_resource):
        if not expected_resource:
            # Some infra resources do not have the monitored
            # attribute, but are treated as if they are monitored.
            if not getattr(actual_resource, 'monitored', True):
                self._handle_unexpected_aim_resource(actual_resource)
        else:
            # Update both monitored and unmonitored
            # resources. Monitored resources should have started out
            # as copies of actual resources, but may have had changes
            # made, such as additional contracts provided or consumed.
            if not expected_resource.user_equal(actual_resource):
                self._handle_incorrect_aim_resource(
                    expected_resource, actual_resource)

    def _handle_unexpected_aim_resource(self, actual_resource):
        if self.should_repair(
                "unexpected %(type)s: %(actual)r" %
                {'type': actual_resource._aci_mo_name,
                 'actual': actual_resource},
                "Deleting"):
            self.aim_mgr.delete(self.actual_aim_ctx, actual_resource)

    def _handle_incorrect_aim_resource(self, expected_resource,
                                       actual_resource):
        if self.should_repair(
                "incorrect %(type)s: %(actual)r which should be: "
                "%(expected)r" %
                {'type': expected_resource._aci_mo_name,
                 'actual': actual_resource,
                 'expected': expected_resource}):
            self.aim_mgr.create(
                self.actual_aim_ctx, expected_resource, overwrite=True)

    def _handle_missing_aim_resource(self, expected_resource):
        if self.should_repair(
                "missing %(type)s: %(expected)r" %
                {'type': expected_resource._aci_mo_name,
                 'expected': expected_resource}):
            self.aim_mgr.create(self.actual_aim_ctx, expected_resource)

    def _validate_db_instances(self):
        for db_class in self._expected_db_instances.keys():
            self._validate_db_instance_class(db_class)

    def _validate_db_instance_class(self, db_class):
        expected_instances = self._expected_db_instances[db_class]
        actual_instances = self.actual_session.query(db_class).all()

        for actual_instance in actual_instances:
            self._validate_actual_db_instance(
                actual_instance, expected_instances)

        for expected_instance in expected_instances.values():
            self._handle_missing_db_instance(expected_instance)

    def _validate_actual_db_instance(self, actual_instance,
                                     expected_instances):
        primary_keys = self._db_instance_primary_keys[
            actual_instance.__class__]
        key = tuple([getattr(actual_instance, k) for k in primary_keys])
        expected_instance = expected_instances.pop(key, None)
        if not expected_instance:
            self._handle_unexpected_db_instance(actual_instance)
        else:
            if not self._is_db_instance_correct(
                    expected_instance, actual_instance):
                self._handle_incorrect_db_instance(
                    expected_instance, actual_instance)

    def _is_db_instance_correct(self, expected_instance, actual_instance):
        expected_values = expected_instance.__dict__
        actual_values = actual_instance.__dict__
        return all([v == actual_values[k] for k, v in expected_values.items()
                    if not k.startswith('_')])

    def _handle_unexpected_db_instance(self, actual_instance):
        if self.should_repair(
                "unexpected %(type)s record: %(actual)s" %
                {'type': actual_instance.__tablename__,
                 'actual': actual_instance.__dict__},
                "Deleting"):
            self.actual_session.delete(actual_instance)

    def _handle_incorrect_db_instance(self, expected_instance,
                                      actual_instance):
        if self.should_repair(
                "incorrect %(type)s record: %(actual)s which should be: "
                "%(expected)s" %
                {'type': expected_instance.__tablename__,
                 'actual': actual_instance.__dict__,
                 'expected': expected_instance.__dict__}):
            self.actual_session.merge(expected_instance)

    def _handle_missing_db_instance(self, expected_instance):
        if self.should_repair(
                "missing %(type)s record: %(expected)s" %
                {'type': expected_instance.__tablename__,
                 'expected': expected_instance.__dict__}):
            self.actual_session.add(expected_instance)


class ValidationAimStore(aim_store.AimStore):

    def __init__(self, validation_mgr):
        self._mgr = validation_mgr
        self.db_session = validation_mgr.expected_session

    def add(self, db_obj):
        self._mgr.expect_aim_resource(db_obj, True)

    def delete(self, db_obj):
        self._mgr.expect_aim_resource(db_obj, remove=True)

    def query(self, db_obj_type, resource_class, in_=None, notin_=None,
              order_by=None, lock_update=False, **filters):
        assert(in_ is None)
        assert(notin_ is None)
        assert(order_by is None)
        if filters:
            if (set(filters.keys()) ==
                set(resource_class.identity_attributes.keys())):
                identity = resource_class(**filters)
                resource = self._mgr.expected_aim_resource(identity)
                return [resource] if resource else []
            else:
                return [r for r in
                        self._mgr.expected_aim_resources(resource_class)
                        if all([getattr(r, k) == v for k, v in
                                filters.items()])]
        else:
            return self._mgr.expected_aim_resources(resource_class)

    def count(self, db_obj_type, resource_class, in_=None, notin_=None,
              **filters):
        assert(False)

    def delete_all(self, db_obj_type, resource_class, in_=None, notin_=None,
                   **filters):
        assert(False)

    def from_attr(self, db_obj, resource_class, attribute_dict):
        for k, v in attribute_dict.items():
            setattr(db_obj, k, v)

    def to_attr(self, resource_class, db_obj):
        assert(False)

    def make_resource(self, cls, db_obj, include_aim_id=False):
        return copy.deepcopy(db_obj)

    def make_db_obj(self, resource):
        result = copy.deepcopy(resource)
        if isinstance(result, aim_resource.EndpointGroup):
            # Since aim.db.models.EndpointGroup.to_attr() maintains
            # openstack_vmm_domain_names for backward compatibility,
            # we do so here.
            result.openstack_vmm_domain_names = [d['name'] for d in
                                                 result.vmm_domains
                                                 if d['type'] == 'OpenStack']
        return result


@contextmanager
def _begin():
    yield


class ValidationSession(object):
    # This is a very minimal implementation of a sqlalchemy DB session
    # (and query), providing only the functionality needed to simulate
    # and validate DB usage buried within library code that cannot be
    # otherwise validated. If more functionality is needed, consider
    # using a sqlite-backed sqlalchemy session instead.

    def __init__(self, validation_mgr):
        self._mgr = validation_mgr

    def begin(self, subtransactions=False, nested=False):
        return _begin()

    def add(self, instance):
        self._mgr.expect_db_instance(instance)

    def query(self, *entities, **kwargs):
        return ValidationQuery(self._mgr, entities, kwargs)


class ValidationQuery(object):

    def __init__(self, validation_mgr, entities, args):
        self._mgr = validation_mgr
        self._entities = entities
        self._args = args
        self._filters = {}

    def filter_by(self, **kwargs):
        self._filters.update(kwargs)
        return self

    def all(self):
        return self._mgr.query_db_instances(
            self._entities, self._args, self._filters)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


BASE_CONTROLLER = 'base_controller'
REFERENCE_CONTROLLER = 'reference_controller'
ADVANCED_CONTROLLER = 'advanced_controller'

base_with_vm = 'base_with_vm'
base = 'base'
advanced = 'advanced'
modes = [base, base_with_vm, advanced]

controller_mode_map = {
        base: BASE_CONTROLLER,
        base_with_vm: REFERENCE_CONTROLLER,
        advanced: ADVANCED_CONTROLLER
}

controllers = {
    controller_mode_map[base]: 'gbpservice.nfp.base_configurator.controllers',
    controller_mode_map[base_with_vm]: ('gbpservice.contrib'
                           '.nfp_service.reference_configurator.controllers'),
    controller_mode_map[advanced]: ('gbpservice.contrib.nfp.configurator'
                          '.advanced_controller.controller_loader')
}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import os

import six

from gbpservice.contrib.nfp.configurator.agents import agent_base
from gbpservice.contrib.nfp.configurator.lib import (
    generic_config_constants as gen_cfg_const)
from gbpservice.contrib.nfp.configurator.lib import constants as common_const
from gbpservice.contrib.nfp.configurator.lib import data_parser
from gbpservice.contrib.nfp.configurator.lib import utils
from gbpservice.nfp.core import event as nfp_event
from gbpservice.nfp.core import log as nfp_logging
from gbpservice.nfp.core import module as nfp_api

LOG = nfp_logging.getLogger(__name__)


class GenericConfigRpcManager(agent_base.AgentBaseRPCManager):
    """Implements APIs invoked by configurator for processing RPC messages.

    RPC client of configurator module receives RPC messages from REST server
    and invokes the API of this class. The instance of this class is registered
    with configurator module using register_service_agent API. Configurator
    module identifies the service agent object based on service type and
    invokes ones of the methods of this class to configure the device.

    """

    def __init__(self, sc, conf):
        """Instantiates child and parent class objects.

        Passes the instances of core service controller and oslo configuration
        to parent instance inorder to provide event enqueue facility for batch
        processing event.

        :param sc: Service Controller object that is used for interfacing
        with core service controller.
        :param conf: Configuration object that is used for configuration
        parameter access.

        """

        self.parse = data_parser.DataParser()
        super(GenericConfigRpcManager, self).__init__(sc, conf)

    def _send_event(self, context, resource_data, event_id, event_key=None):
        """Posts an event to framework.

        :param context: The agent info dictionary prepared in demuxer library
         which contains the API context alongside other information.
        :param kwargs: Keyword arguments which are passed as data to event
        :param event_id: Unique identifier for the event
        :param event_key: Event key for serialization

        """

        arg_dict = {'context': context,
                    'resource_data': resource_data}
        ev = self.sc.new_event(id=event_id, data=arg_dict, key=event_key)
        self.sc.post_event(ev)

    def configure_interfaces(self, context, resource_data):
        """Enqueues event for worker to process configure interfaces request.

        :param context: The agent info dictionary prepared in demuxer library
         which contains the API context alongside other information.
        :param kwargs: RPC Request data

        Returns: None

        """

        self._send_event(context,
                         resource_data,
                         gen_cfg_const.EVENT_CONFIGURE_INTERFACES)

    def clear_interfaces(self, context, resource_data):
        """Enqueues event for worker to process clear interfaces request.

        :param context: The agent info dictionary prepared in demuxer library
         which contains the API context alongside other information.
        :param kwargs: RPC Request data

        Returns: None

        """

        self._send_event(context,
                         resource_data,
                         gen_cfg_const.EVENT_CLEAR_INTERFACES)

    def configure_routes(self, context, resource_data):
        """Enqueues event for worker to process configure routes request.

        :param context: The agent info dictionary prepared in demuxer library
         which contains the API context alongside other information.
        :param kwargs: RPC Request data

        Returns: None

        """

        self._send_event(context,
                         resource_data,
                         gen_cfg_const.EVENT_CONFIGURE_ROUTES)

    def clear_routes(self, context, resource_data):
        """Enqueues event for worker to process clear routes request.

        :param context: The agent info dictionary prepared in demuxer library
         which contains the API context alongside other information.
        :param kwargs: RPC Request data

        Returns: None

        """

        self._send_event(context,
                         resource_data,
                         gen_cfg_const.EVENT_CLEAR_ROUTES)

    def configure_healthmonitor(self, context, resource_data):
        """Enqueues event for worker to process configure healthmonitor
        request.

        :param context: The agent info dictionary prepared in demuxer library
         which contains the API context alongside other information.
        :param kwargs: RPC Request data

        Returns: None

        """
        LOG.info("Received configure health monitor api for nfds:"
                 "%(nfds)s",
                 {'nfds': resource_data['nfds']})
        resource_data['fail_count'] = 0
        self._send_event(context,
                         resource_data,
                         gen_cfg_const.EVENT_CONFIGURE_HEALTHMONITOR,
                         resource_data['nfds'][0]['vmid'])

    def clear_healthmonitor(self, context, resource_data):
        """Enqueues event for worker to process clear healthmonitor request.

        :param context: The agent info dictionary prepared in demuxer library
         which contains the API context alongside other information.
        :param kwargs: RPC Request data

        Returns: None

        """
        LOG.info("Received clear health monitor api for nfds:"
                 "%(nfds)s",
                 {'nfds': resource_data['nfds']})
        event_key = resource_data['nfds'][0]['vmid']
        poll_event_id = gen_cfg_const.EVENT_CONFIGURE_HEALTHMONITOR
        self.sc.stop_poll_event(event_key, poll_event_id)


class GenericConfigEventHandler(agent_base.AgentBaseEventHandler,
                                nfp_api.NfpEventHandler):
    """Implements event handlers and their helper methods.

    Object of this class is registered with the event class of core service
    controller. Based on the event key, handle_event method of this class is
    invoked by core service controller.
    """

    def __init__(self, sc, drivers, rpcmgr):
        super(GenericConfigEventHandler, self).__init__(
            sc, drivers, rpcmgr)
        self.sc = sc

    def _get_driver(self, service_type, service_vendor, service_feature):
        """Retrieves service driver object based on service type input.

        Currently, service drivers are identified with service type. Support
        for single driver per service type is provided. When multi-vendor
        support is going to be provided, the driver should be selected based
        on both service type and vendor name.

        :param service_type: Service type - firewall/vpn/loadbalancer

        Returns: Service driver instance

        """

        return self.drivers[service_type + service_vendor + service_feature]

    def handle_event(self, ev):
        """Processes the generated events in worker context.

        Processes the following events.
        - Configure Interfaces
        - Clear Interfaces
        - Configure routes
        - Clear routes
        - Configure health monitor
        - Clear health monitor
        Enqueues responses into notification queue.

        Returns: None

        """
        try:
            event_data = ev.data
            if ev.id == 'PROCESS_BATCH':
                NFI = event_data['sa_req_list'][0][
                    'agent_info']['context']['nfi_id']
                NF = event_data['sa_req_list'][0][
                    'agent_info']['context']['nf_id']
            else:
                NFI = event_data['context']['context']['nfi_id']
                NF = event_data['context']['context']['nf_id']
        except Exception:
            NFI = None
            NF = None

        msg = ("Handling event '%s', with NF:%s and NFI:%s"
               % (ev.id, NF, NFI))
        LOG.info(msg)

        # Process batch of request data blobs
        try:
            # Process batch of request data blobs
            if ev.id == common_const.EVENT_PROCESS_BATCH:
                self.process_batch(ev)
                return
            # Process HM poll events
            elif ev.id == gen_cfg_const.EVENT_CONFIGURE_HEALTHMONITOR:
                resource_data = ev.data.get('resource_data')
                periodicity = resource_data['nfds'][0]['periodicity']
                EV_CONF_HM_MAXRETRY = (
                    gen_cfg_const.EVENT_CONFIGURE_HEALTHMONITOR_MAXRETRY)
                if periodicity == gen_cfg_const.INITIAL:
                    self.sc.poll_event(
                        ev,
                        max_times=EV_CONF_HM_MAXRETRY)

                elif periodicity == gen_cfg_const.FOREVER:
                    self.sc.poll_event(ev)
            else:
                self._process_event(ev)
        except Exception as err:
            msg = ("Failed to process event %s, reason %s "
                   % (ev.data, err))
            LOG.error(msg)
            return

    def send_periodic_hm_notification(self, ev, nfd, result, notification_id):
        ev_copy = copy.deepcopy(ev)
        ev_copy.data["context"]["notification_data"] = {}
        ev_copy.data["context"]["context"]["nfp_context"]["id"] = (
            notification_id)
        ev_copy.data['context']['context']['nfd_id'] = nfd.get('vmid')
        notification_data = self._prepare_notification_data(ev_copy, result)
        self.notify._notification(notification_data)

    def handle_periodic_hm(self, ev, result):
        resource_data = ev.data['resource_data']
        nfd = ev.data["resource_data"]['nfds'][0]
        periodic_polling_reason = nfd["periodic_polling_reason"]
        if result == common_const.FAILED:
            """If health monitoring fails continuously for MAX_FAIL_COUNT times
                send fail notification to orchestrator
            """
            resource_data['fail_count'] = resource_data.get('fail_count') + 1
            if (resource_data.get('fail_count') >=
                    gen_cfg_const.MAX_FAIL_COUNT):
                # REVISIT(Shishir): Remove statefull logic from here,
                # need to come up with statleless logic.
                if periodic_polling_reason == (
                        gen_cfg_const.DEVICE_TO_BECOME_DOWN):
                    notification_id = gen_cfg_const.DEVICE_NOT_REACHABLE
                    self.send_periodic_hm_notification(ev, nfd, result,
                                                       notification_id)
                    nfd["periodic_polling_reason"] = (
                        gen_cfg_const.DEVICE_TO_BECOME_UP)
        elif result == common_const.SUCCESS:
            """set fail_count to 0 if it had failed earlier even once
            """
            resource_data['fail_count'] = 0
            if periodic_polling_reason == gen_cfg_const.DEVICE_TO_BECOME_UP:
                notification_id = gen_cfg_const.DEVICE_REACHABLE
                self.send_periodic_hm_notification(ev, nfd, result,
                                                   notification_id)
                nfd["periodic_polling_reason"] = (
                    gen_cfg_const.DEVICE_TO_BECOME_DOWN)

    def _process_event(self, ev):
        LOG.debug(" Handling event %s ", (ev.data))
        # Process single request data blob
        resource_data = ev.data['resource_data']
        # The context inside ev.data is the agent info dictionary prepared
        # in demuxer library which contains the API context alongside
        # other information like service vendor, type etc..
        agent_info = ev.data['context']
        context = agent_info['context']
        service_type = agent_info['resource_type']
        service_vendor = agent_info['service_vendor']
        service_feature = agent_info.get('service_feature', '')

        try:
            msg = ("Worker process with ID: %s starting "
                   "to handle task: %s for service type: %s. "
                   % (os.getpid(), ev.id, str(service_type)))
            LOG.debug(msg)

            driver = self._get_driver(service_type, service_vendor,
                                      service_feature)

            # Invoke service driver methods based on event type received
            result = getattr(driver, "%s" % ev.id.lower())(context,
                                                           resource_data)
        except Exception as err:
            msg = ("Failed to process ev.id=%s, ev=%s reason=%s" %
                   (ev.id, ev.data, err))
            LOG.error(msg)
            result = common_const.FAILED

        if ev.id == gen_cfg_const.EVENT_CONFIGURE_HEALTHMONITOR:
            if (resource_data['nfds'][0][
                'periodicity'] == gen_cfg_const.INITIAL and
                    result == common_const.SUCCESS):
                notification_data = self._prepare_notification_data(ev,
                                                                    result)
                self.notify._notification(notification_data)
                msg = ("VM Health check successful")
                LOG.info(msg)
                return {'poll': False}
            elif resource_data['nfds'][0]['periodicity'] == (
                    gen_cfg_const.FOREVER):
                ev.data["context"]["resource"] = gen_cfg_const.PERIODIC_HM
                self.handle_periodic_hm(ev, result)
        else:
            """For other events, irrespective of result send notification"""
            notification_data = self._prepare_notification_data(ev, result)
            self.notify._notification(notification_data)

    def prepare_notification_result(self, result):
        if result in common_const.SUCCESS:
            data = {'status_code': common_const.SUCCESS}
        else:
            data = {'status_code': common_const.FAILURE,
                    'error_msg': result}
        return data

    def _prepare_notification_data(self, ev, result):
        """Prepare notification data as expected by config agent

        :param ev: event object
        :param result: result of the handled event

        Returns: notification_data

        """
        agent_info = ev.data['context']
        context = agent_info['context']

        # Retrieve notification and remove it from context. Context is used
        # as transport from batch processing function to this last event
        # processing function. To keep the context unchanged, delete the
        # notification_data before invoking driver API.
        notification_data = agent_info['notification_data']
        service_type = agent_info['resource_type']
        resource = agent_info['resource']

        data = self.prepare_notification_result(result)
        msg = {'info': {'service_type': service_type,
                        'context': context},
               'notification': [{'resource': resource,
                                 'data': data}]
               }
        if not notification_data:
            notification_data.update(msg)
        else:
            data = {'resource': resource,
                    'data': data}
            notification_data['notification'].append(data)
        return notification_data

    def event_cancelled(self, ev, reason):
        """Invoked by process framework when poll ev object reaches
           polling threshold ev.max_times.
           Finally it Enqueues response into notification queue.

        :param ev: Event object

        Returns: None

        """
        msg = ('Cancelled poll event. Event Data: %s ' % (ev.data))
        LOG.error(msg)
        result = common_const.FAILED
        notification_data = self._prepare_notification_data(ev, result)
        self.notify._notification(notification_data)

    @nfp_api.poll_event_desc(
        event=gen_cfg_const.EVENT_CONFIGURE_HEALTHMONITOR,
        spacing=gen_cfg_const.EVENT_CONFIGURE_HEALTHMONITOR_SPACING)
    def handle_configure_healthmonitor(self, ev):
        """Decorator method called for poll event CONFIGURE_HEALTHMONITOR
           Finally it Enqueues response into notification queue.

        :param ev: Event object

        Returns: None

        """
        return self._process_event(ev)


def events_init(sc, drivers, rpcmgr):
    """Registers events with core service controller.

    All the events will come to handle_event method of class instance
    registered in 'handler' field.

    :param drivers: Driver instances registered with the service agent
    :param rpcmgr: Instance to receive all the RPC messages from configurator
    module.

    Returns: None

    """

    event_id_list = [
        gen_cfg_const.EVENT_CONFIGURE_INTERFACES,
        gen_cfg_const.EVENT_CLEAR_INTERFACES,
        gen_cfg_const.EVENT_CONFIGURE_ROUTES,
        gen_cfg_const.EVENT_CLEAR_ROUTES,
        gen_cfg_const.EVENT_CONFIGURE_HEALTHMONITOR,
        gen_cfg_const.EVENT_CLEAR_HEALTHMONITOR,
        common_const.EVENT_PROCESS_BATCH
    ]
    events = []

    for event in event_id_list:
        events.append(
            nfp_event.Event(
                id=event,
                handler=GenericConfigEventHandler(sc, drivers, rpcmgr)))

    sc.register_events(events)


def load_drivers(conf):
    """Imports all the driver files.

    Returns: Dictionary of driver objects with a specified service type and
    vendor name

    """

    cutils = utils.ConfiguratorUtils(conf)
    drivers = cutils.load_drivers()

    for service_type, driver_name in six.iteritems(drivers):
        driver_obj = driver_name(conf=conf)
        drivers[service_type] = driver_obj
    LOG.info("Generic config agent loaded drivers drivers:"
             "%(drivers)s",
             {'drivers': drivers})
    return drivers


def register_service_agent(cm, sc, conf, rpcmgr):
    """Registers generic configuration service agent with configurator module.

    :param cm: Instance of configurator module
    :param sc: Instance of core service controller
    :param conf: Instance of oslo configuration
    :param rpcmgr: Instance containing RPC methods which are invoked by
    configurator module on corresponding RPC message arrival

    """

    service_type = gen_cfg_const.SERVICE_TYPE
    cm.register_service_agent(service_type, rpcmgr)


def init_agent(cm, sc, conf):
    """Initializes generic configuration agent.

    :param cm: Instance of configuration module
    :param sc: Instance of core service controller
    :param conf: Instance of oslo configuration

    """

    try:
        drivers = load_drivers(conf)
    except Exception as err:
        msg = ("Generic configuration agent failed to load service drivers."
               "Error:%s"
               % (str(err).capitalize()))
        LOG.error(msg)
        raise err
    else:
        msg = ("Generic configuration agent loaded service"
               " drivers successfully.")
        LOG.debug(msg)

    rpcmgr = GenericConfigRpcManager(sc, conf)

    try:
        events_init(sc, drivers, rpcmgr)
    except Exception as err:
        msg = ("Generic configuration agent failed to initialize events. %s"
               % (str(err).capitalize()))
        LOG.error(msg)
        raise err
    else:
        msg = ("Generic configuration agent initialized"
               " events successfully.")
        LOG.debug(msg)

    try:
        register_service_agent(cm, sc, conf, rpcmgr)
    except Exception as err:
        msg = ("Failed to register generic configuration agent with"
               " configurator module. %s" % (str(err).capitalize()))
        LOG.error(msg)
        raise err
    else:
        msg = ("Generic configuration agent registered with configuration"
               " module successfully.")
        LOG.debug(msg)


def init_agent_complete(cm, sc, conf):
    msg = ("Initialization of generic configuration agent completed.")
    LOG.info(msg)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_config import cfg as oslo_config

from gbpservice.nfp.common import constants as nfp_constants
from gbpservice.nfp.core import context
from gbpservice.nfp.orchestrator import context as module_context

context.NfpContext = module_context.NfpContext

openstack_opts = [
    oslo_config.StrOpt('auth_host',
                       default='localhost',
                       help='Openstack controller IP Address'),
    # REVISIT: In future, use nfp_user with admin role instead of admin_user
    oslo_config.StrOpt('admin_user',
                       help='Admin user name to create service VMs'),
    oslo_config.StrOpt('admin_password',
                       help='Admin password to create service VMs'),
    # REVISIT: In future, use nfp_tenant_name instead of admin_tenant_name
    oslo_config.StrOpt('admin_tenant_name',
                       help='Admin tenant name to create service VMs'),
    oslo_config.StrOpt('admin_tenant_id',
                       help='Admin tenant ID to create service VMs'),
    oslo_config.StrOpt('auth_protocol',
                       default='http', help='Auth protocol used.'),
    oslo_config.IntOpt('auth_port',
                       default='5000', help='Auth protocol used.'),
    oslo_config.IntOpt('bind_port',
                       default='9696', help='Auth protocol used.'),
    oslo_config.StrOpt('auth_version',
                       default='v2.0', help='Auth protocol used.'),
    oslo_config.StrOpt('auth_uri',
                       default='', help='Auth URI.'),
]

oslo_config.CONF.register_opts(openstack_opts, "nfp_keystone_authtoken")

nfp_orchestrator_opts = [
    oslo_config.ListOpt(
        'supported_vendors',
        default=[nfp_constants.VYOS_VENDOR,
                 nfp_constants.HAPROXY_LBAASV2, nfp_constants.NFP_VENDOR],
        help="Supported service vendors for nfp"),
    oslo_config.StrOpt('monitoring_ptg_l3policy_id',
                       default='')
]

oslo_config.CONF.register_opts(nfp_orchestrator_opts, 'orchestrator')

device_orchestrator_opts = [
    oslo_config.BoolOpt('volume_support',
                        default=False, help='cinder volume support'),
    oslo_config.StrOpt('volume_size',
                       default='2', help='cinder volume size')
]

oslo_config.CONF.register_opts(device_orchestrator_opts, 'device_orchestrator')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ast
import copy

from gbpservice.contrib.nfp.config_orchestrator.common import common
from gbpservice.nfp.common import constants as const
from gbpservice.nfp.common import data_formatter as df
from gbpservice.nfp.common import utils
from gbpservice.nfp.core import context as module_context
from gbpservice.nfp.core import log as nfp_logging
from gbpservice.nfp.lib import transport

from neutron_fwaas.db.firewall import firewall_db

from oslo_log import helpers as log_helpers
import oslo_messaging as messaging

LOG = nfp_logging.getLogger(__name__)

"""
RPC handler for Firewall service
"""


class FwAgent(firewall_db.Firewall_db_mixin):

    RPC_API_VERSION = '1.0'
    target = messaging.Target(version=RPC_API_VERSION)

    def __init__(self, conf, sc):
        super(FwAgent, self).__init__()
        self._conf = conf
        self._sc = sc
        self._db_inst = super(FwAgent, self)

    def _get_firewalls(self, context, tenant_id,
                       firewall_policy_id, description):
        filters = {'tenant_id': [tenant_id],
                   'firewall_policy_id': [firewall_policy_id]}
        args = {'context': context, 'filters': filters}
        firewalls = self._db_inst.get_firewalls(**args)
        for firewall in firewalls:
            firewall['description'] = description
        return firewalls

    def _get_firewall_policies(self, context, tenant_id,
                               firewall_policy_id, description):
        filters = {'tenant_id': [tenant_id],
                   'id': [firewall_policy_id]}
        args = {'context': context, 'filters': filters}
        firewall_policies = self._db_inst.get_firewall_policies(**args)
        return firewall_policies

    def _get_firewall_rules(self, context, tenant_id,
                            firewall_policy_id, description):
        filters = {'tenant_id': [tenant_id],
                   'firewall_policy_id': [firewall_policy_id]}
        args = {'context': context, 'filters': filters}
        firewall_rules = self._db_inst.get_firewall_rules(**args)
        return firewall_rules

    def _get_firewall_context(self, **kwargs):
        firewalls = self._get_firewalls(**kwargs)
        firewall_policies = self._get_firewall_policies(**kwargs)
        firewall_rules = self._get_firewall_rules(**kwargs)
        return {'firewalls': firewalls,
                'firewall_policies': firewall_policies,
                'firewall_rules': firewall_rules}

    def _context(self, **kwargs):
        context = kwargs.get('context')
        if context.is_admin:
            kwargs['tenant_id'] = context.tenant_id
        db = self._get_firewall_context(**kwargs)
        return db

    def _prepare_resource_context_dicts(self, **kwargs):
        # Prepare context_dict
        context = kwargs.get('context')
        context_resource_data = kwargs.pop('context_resource_data')
        ctx_dict = context.to_dict()
        # Collecting db entry required by configurator.
        # Addind service_info to neutron context and sending
        # dictionary format to the configurator.
        db = self._context(**kwargs)
        rsrc_ctx_dict = copy.deepcopy(ctx_dict)
        rsrc_ctx_dict.update({'service_info': db})
        rsrc_ctx_dict.update({'resource_data': context_resource_data})
        return ctx_dict, rsrc_ctx_dict

    def _get_resource_data(self, description, resource_type):
        resource_data = df.get_network_function_info(description,
                                                     resource_type)
        return resource_data

    def _update_request_data(self, body, description):
        pass

    def _data_wrapper(self, context, firewall, host, nf, reason):
        # Hardcoding the position for fetching data since we are owning
        # its positional change
        description = ast.literal_eval((nf['description'].split('\n'))[1])
        description.update({'tenant_id': firewall['tenant_id']})
        context_resource_data = self._get_resource_data(description,
                                                        const.FIREWALL)
        fw_mac = description['provider_ptg_info'][0]
        # REVISIT(dpak): We need to avoid resource description
        # dependency in OTC and instead use neutron context description.
        firewall.update({'description': str(description)})
        kwargs = {'context': context,
                  'context_resource_data': context_resource_data,
                  'firewall_policy_id': firewall[
                      'firewall_policy_id'],
                  'description': str(description),
                  'tenant_id': firewall['tenant_id']}

        ctx_dict, rsrc_ctx_dict = self._prepare_resource_context_dicts(
            **kwargs)
        service_vm_context = utils.get_service_vm_context(
            description['service_vendor'])
        nfp_context = {'network_function_id': nf['id'],
                       'neutron_context': ctx_dict,
                       'fw_mac': fw_mac,
                       'requester': 'nas_service',
                       'logging_context': module_context.get()['log_context'],
                       'service_vm_context': service_vm_context}
        resource = resource_type = 'firewall'
        resource_data = {resource: firewall,
                         'host': host,
                         'neutron_context': rsrc_ctx_dict}
        body = common.prepare_request_data(nfp_context, resource,
                                           resource_type, resource_data,
                                           description['service_vendor'])
        self._update_request_data(body, description)
        return body

    def _fetch_nf_from_resource_desc(self, desc):
        desc_dict = ast.literal_eval(desc)
        nf_id = desc_dict['network_function_id']
        return nf_id

    @log_helpers.log_method_call
    def create_firewall(self, context, firewall, host):
        nfp_context = module_context.init()
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(firewall["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        LOG.info("Received RPC CREATE FIREWALL for "
                 "Firewall: %(firewall)s",
                 {'firewall': firewall})
        body = self._data_wrapper(context, firewall, host, nf, 'CREATE')
        transport.send_request_to_configurator(self._conf,
                                               context, body, "CREATE")

    @log_helpers.log_method_call
    def delete_firewall(self, context, firewall, host):
        nfp_context = module_context.init()
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(firewall["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        LOG.info("Received RPC DELETE FIREWALL for "
                 "Firewall: %(firewall)s",
                 {'firewall': firewall})
        body = self._data_wrapper(context, firewall, host, nf, 'DELETE')
        transport.send_request_to_configurator(self._conf,
                                               context, body, "DELETE")



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""ncp_node_instance_stacks

Revision ID: 5358a28fb97d
Revises: d08627f64e37

"""

# revision identifiers, used by Alembic.
revision = '5358a28fb97d'
down_revision = 'd08627f64e37'


from alembic import op
import sqlalchemy as sa


def upgrade(active_plugins=None, options=None):
    op.create_table(
        'ncp_node_instance_stacks',
        sa.Column('sc_instance_id', sa.String(length=36), nullable=False),
        sa.Column('sc_node_id', sa.String(length=36), nullable=False),
        sa.Column('stack_id', sa.String(length=36), nullable=False),
        sa.PrimaryKeyConstraint('sc_instance_id', 'sc_node_id', 'stack_id')
    )


def downgrade(active_plugins=None, options=None):
    pass



# Copyright (c) 2017 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from aim.api import resource as aim_resource
from neutron.db.models import address_scope as as_db
from neutron.db import models_v2
from neutron_lib import context as n_context
from neutron_lib.db import model_base
from oslo_db import exception as db_exc
from oslo_log import log
import sqlalchemy as sa
from sqlalchemy.ext import baked
from sqlalchemy import orm

from gbpservice.neutron.db import api as db_api

VM_UPDATE_PURPOSE = 'VmUpdate'

LOG = log.getLogger(__name__)

BAKERY = baked.bakery(_size_alert=lambda c: LOG.warning(
    "sqlalchemy baked query cache size exceeded in %s", __name__))


class AddressScopeMapping(model_base.BASEV2):
    __tablename__ = 'apic_aim_address_scope_mappings'

    scope_id = sa.Column(
        sa.String(36), sa.ForeignKey('address_scopes.id', ondelete='CASCADE'),
        primary_key=True)

    address_scope = orm.relationship(
        as_db.AddressScope,
        backref=orm.backref(
            'aim_mapping', lazy='joined', uselist=False, cascade='delete'))

    vrf_name = sa.Column(sa.String(64))
    vrf_tenant_name = sa.Column(sa.String(64))
    vrf_owned = sa.Column(sa.Boolean, nullable=False)


class NetworkMapping(model_base.BASEV2):
    __tablename__ = 'apic_aim_network_mappings'

    network_id = sa.Column(
        sa.String(36), sa.ForeignKey('networks.id', ondelete='CASCADE'),
        primary_key=True)

    network = orm.relationship(
        # REVISIT: The following also had an eager loading using
        # lazy='joined' which was removed since it presumably causes
        # cascaded delete to be triggered twice on the NetworkSegment
        # object. The issue can be reproduced with the following UT:
        # gbpservice.neutron.tests.unit.plugins.ml2plus.test_apic_aim.
        # TestAimMapping.test_network_lifecycle
        # Removing the eager loading will cause a performance hit.
        models_v2.Network,
        backref=orm.backref(
            'aim_mapping', lazy='joined', uselist=False, cascade='delete'))

    bd_name = sa.Column(sa.String(64))
    bd_tenant_name = sa.Column(sa.String(64))

    epg_name = sa.Column(sa.String(64))
    epg_app_profile_name = sa.Column(sa.String(64))
    epg_tenant_name = sa.Column(sa.String(64))

    l3out_name = sa.Column(sa.String(64))
    l3out_ext_net_name = sa.Column(sa.String(64))
    l3out_tenant_name = sa.Column(sa.String(64))

    vrf_name = sa.Column(sa.String(64))
    vrf_tenant_name = sa.Column(sa.String(64))


class HAIPAddressToPortAssociation(model_base.BASEV2):

    """Port Owner for HA IP Address.

    This table is used to store the mapping between the HA IP Address
    and the Port ID of the Neutron Port which currently owns this
    IP Address.
    """

    __tablename__ = 'apic_ml2_ha_ipaddress_to_port_owner'

    ha_ip_address = sa.Column(sa.String(64), nullable=False,
                              primary_key=True)
    port_id = sa.Column(sa.String(64), sa.ForeignKey('ports.id',
                                                     ondelete='CASCADE'),
                        nullable=False, primary_key=True)


class VMName(model_base.BASEV2):
    __tablename__ = 'apic_aim_vm_names'

    device_id = sa.Column(sa.String(255), primary_key=True)
    vm_name = sa.Column(sa.String(255))


# At any point of time, there should only be one entry in this table.
# We will enforce that by using the same value for the purpose column which
# is the primary key. That entry is used to make sure only one controller is
# actively updating the VMName table.
class VMNameUpdate(model_base.BASEV2):
    __tablename__ = 'apic_aim_vm_name_updates'

    purpose = sa.Column(sa.String(36), primary_key=True)
    host_id = sa.Column(sa.String(255))
    last_incremental_update_time = sa.Column(sa.DateTime)
    last_full_update_time = sa.Column(sa.DateTime)


class DbMixin(object):

    # AddressScopeMapping functions.

    def _add_address_scope_mapping(self, session, scope_id, vrf,
                                   vrf_owned=True, update_scope=True):
        mapping = AddressScopeMapping(
            scope_id=scope_id,
            vrf_name=vrf.name,
            vrf_tenant_name=vrf.tenant_name,
            vrf_owned=vrf_owned)
        session.add(mapping)
        if update_scope:
            # The AddressScope instance should already be in the
            # session cache, so this should not add another DB
            # roundtrip. It needs to be updated in case something
            # within the same transaction tries to access its
            # aim_mapping relationship after retrieving the
            # AddressScope record from the session cache.
            query = BAKERY(lambda s: s.query(
                as_db.AddressScope))
            query += lambda q: q.filter_by(
                id=sa.bindparam('scope_id'))
            scope = query(session).params(
                scope_id=scope_id).one_or_none()

            scope.aim_mapping = mapping
        return mapping

    def _get_address_scope_mapping(self, session, scope_id):
        query = BAKERY(lambda s: s.query(
            AddressScopeMapping))
        query += lambda q: q.filter_by(
            scope_id=sa.bindparam('scope_id'))
        return query(session).params(
            scope_id=scope_id).one_or_none()

    def _get_address_scope_mappings_for_vrf(self, session, vrf):
        query = BAKERY(lambda s: s.query(
            AddressScopeMapping))
        query += lambda q: q.filter_by(
            vrf_tenant_name=sa.bindparam('tenant_name'),
            vrf_name=sa.bindparam('name'))
        return query(session).params(
            tenant_name=vrf.tenant_name,
            name=vrf.name).all()

    def _get_address_scopes_owning_vrf(self, session, vrf):
        query = BAKERY(lambda s: s.query(
            as_db.AddressScope))
        query += lambda q: q.join(
            AddressScopeMapping,
            AddressScopeMapping.scope_id == as_db.AddressScope.id)
        query += lambda q: q.filter(
            AddressScopeMapping.vrf_tenant_name == sa.bindparam('tenant_name'),
            AddressScopeMapping.vrf_name == sa.bindparam('name'),
            AddressScopeMapping.vrf_owned)
        query += lambda q: q.order_by(
            as_db.AddressScope.ip_version)
        return query(session).params(
            tenant_name=vrf.tenant_name,
            name=vrf.name).all()

    def _get_address_scope_vrf(self, mapping):
        return aim_resource.VRF(
            tenant_name=mapping.vrf_tenant_name,
            name=mapping.vrf_name)

    # NetworkMapping functions.

    def _add_network_mapping(self, session, network_id, bd, epg, vrf,
                             ext_net=None, update_network=True):
        if not ext_net:
            mapping = NetworkMapping(
                network_id=network_id,
                bd_name=bd.name,
                bd_tenant_name=bd.tenant_name,
                epg_name=epg.name,
                epg_app_profile_name=epg.app_profile_name,
                epg_tenant_name=epg.tenant_name,
                vrf_name=vrf.name,
                vrf_tenant_name=vrf.tenant_name)
        else:
            mapping = NetworkMapping(
                network_id=network_id,
                l3out_name=ext_net.l3out_name,
                l3out_ext_net_name=ext_net.name,
                l3out_tenant_name=ext_net.tenant_name,
                vrf_name=vrf.name,
                vrf_tenant_name=vrf.tenant_name)
        session.add(mapping)
        if update_network:
            # The Network instance should already be in the session
            # cache, so this should not add another DB roundtrip. It
            # needs to be updated in case something within the same
            # transaction tries to access its aim_mapping relationship
            # after retrieving the Network record from the session
            # cache.
            query = BAKERY(lambda s: s.query(
                models_v2.Network))
            query += lambda q: q.filter_by(
                id=sa.bindparam('network_id'))
            net = query(session).params(
                network_id=network_id).one_or_none()

            net.aim_mapping = mapping
        return mapping

    def _get_network_mapping(self, session, network_id):
        query = BAKERY(lambda s: s.query(
            NetworkMapping))
        query += lambda q: q.filter_by(
            network_id=sa.bindparam('network_id'))
        return query(session).params(
            network_id=network_id).one_or_none()

    def _get_network_mappings_for_vrf(self, session, vrf):
        query = BAKERY(lambda s: s.query(
            NetworkMapping))
        query += lambda q: q.filter_by(
            vrf_tenant_name=sa.bindparam('vrf_tenant_name'),
            vrf_name=sa.bindparam('vrf_name'))
        return query(session).params(
            vrf_tenant_name=vrf.tenant_name,
            vrf_name=vrf.name).all()

    def _is_vrf_used_by_networks(self, session, vrf):
        query = BAKERY(lambda s: s.query(
            NetworkMapping.network_id))
        query += lambda q: q.filter_by(
            vrf_tenant_name=sa.bindparam('vrf_tenant_name'),
            vrf_name=sa.bindparam('vrf_name'))
        return query(session).params(
            vrf_tenant_name=vrf.tenant_name,
            vrf_name=vrf.name).first() is not None

    def _get_network_bd(self, mapping):
        return aim_resource.BridgeDomain(
            tenant_name=mapping.bd_tenant_name,
            name=mapping.bd_name)

    def _get_network_epg(self, mapping):
        return aim_resource.EndpointGroup(
            tenant_name=mapping.epg_tenant_name,
            app_profile_name=mapping.epg_app_profile_name,
            name=mapping.epg_name)

    def _get_network_l3out(self, mapping):
        if not mapping:
            # REVISIT: Is this still needed now that
            # _add_network_mapping updates the Network instance's
            # aim_mapping? If so, the test should probably be moved to
            # the caller to make all these
            # _get_<neutron-resource>_<aim-resource> methods more
            # consistent.
            return None
        return aim_resource.L3Outside(
            tenant_name=mapping.l3out_tenant_name,
            name=mapping.l3out_name)

    def _get_network_l3out_ext_net(self, mapping):
        return aim_resource.ExternalNetwork(
            tenant_name=mapping.l3out_tenant_name,
            l3out_name=mapping.l3out_name, name=mapping.l3out_ext_net_name)

    def _get_network_l3out_default_ext_subnetv4(self, mapping):
        return aim_resource.ExternalSubnet(
            tenant_name=mapping.l3out_tenant_name,
            l3out_name=mapping.l3out_name,
            external_network_name=mapping.l3out_ext_net_name,
            cidr="0.0.0.0/0")

    def _get_network_l3out_default_ext_subnetv6(self, mapping):
        return aim_resource.ExternalSubnet(
            tenant_name=mapping.l3out_tenant_name,
            l3out_name=mapping.l3out_name,
            external_network_name=mapping.l3out_ext_net_name,
            cidr="::/0")

    def _get_network_vrf(self, mapping):
        return aim_resource.VRF(
            tenant_name=mapping.vrf_tenant_name,
            name=mapping.vrf_name)

    def _set_network_bd(self, mapping, bd):
        mapping.bd_tenant_name = bd.tenant_name
        mapping.bd_name = bd.name

    def _set_network_epg(self, mapping, epg):
        mapping.epg_tenant_name = epg.tenant_name
        mapping.epg_app_profile_name = epg.app_profile_name
        mapping.epg_name = epg.name

    def _set_network_l3out(self, mapping, l3out):
        mapping.l3out_tenant_name = l3out.tenant_name
        mapping.l3out_name = l3out.name

    def _set_network_vrf(self, mapping, vrf):
        mapping.vrf_tenant_name = vrf.tenant_name
        mapping.vrf_name = vrf.name

    # HAIPAddressToPortAssociation functions.

    def _get_ha_ipaddress(self, port_id, ipaddress, session=None):
        session = session or db_api.get_reader_session()

        query = BAKERY(lambda s: s.query(
            HAIPAddressToPortAssociation))
        query += lambda q: q.filter_by(
            port_id=sa.bindparam('port_id'),
            ha_ip_address=sa.bindparam('ipaddress'))
        return query(session).params(
            port_id=port_id, ipaddress=ipaddress).first()

    def get_port_for_ha_ipaddress(self, ipaddress, network_id,
                                  session=None):
        """Returns the Neutron Port ID for the HA IP Addresss."""
        session = session or db_api.get_reader_session()
        query = BAKERY(lambda s: s.query(
            HAIPAddressToPortAssociation))
        query += lambda q: q.join(
            models_v2.Port,
            models_v2.Port.id == HAIPAddressToPortAssociation.port_id)
        query += lambda q: q.filter(
            HAIPAddressToPortAssociation.ha_ip_address ==
            sa.bindparam('ipaddress'))
        query += lambda q: q.filter(
            models_v2.Port.network_id == sa.bindparam('network_id'))
        port_ha_ip = query(session).params(
            ipaddress=ipaddress, network_id=network_id).first()
        return port_ha_ip

    def get_ha_ipaddresses_for_port(self, port_id, session=None):
        """Returns the HA IP Addressses associated with a Port."""
        session = session or db_api.get_reader_session()

        query = BAKERY(lambda s: s.query(
            HAIPAddressToPortAssociation))
        query += lambda q: q.filter_by(
            port_id=sa.bindparam('port_id'))
        objs = query(session).params(
            port_id=port_id).all()

        # REVISIT: Do the sorting in the UT?
        return sorted([x['ha_ip_address'] for x in objs])

    def set_port_id_for_ha_ipaddress(self, port_id, ipaddress, session=None):
        """Stores a Neutron Port Id as owner of HA IP Addr (idempotent API)."""
        session = session or db_api.get_writer_session()
        try:
            with session.begin(subtransactions=True):
                obj = self._get_ha_ipaddress(port_id, ipaddress, session)
                if obj:
                    return obj
                else:
                    obj = HAIPAddressToPortAssociation(
                        port_id=port_id, ha_ip_address=ipaddress)
                    session.add(obj)
                    return obj
        except db_exc.DBDuplicateEntry:
            LOG.debug('Duplicate IP ownership entry for tuple %s',
                      (port_id, ipaddress))

    def delete_port_id_for_ha_ipaddress(self, port_id, ipaddress,
                                        session=None):
        session = session or db_api.get_writer_session()
        with session.begin(subtransactions=True):
            try:
                # REVISIT: Can this query be baked? The
                # sqlalchemy.ext.baked.Result class does not have a
                # delete() method, and adding delete() to the baked
                # query before executing it seems to result in the
                # params() not being evaluated.
                return session.query(
                    HAIPAddressToPortAssociation).filter_by(
                        port_id=port_id,
                        ha_ip_address=ipaddress).delete()
            except orm.exc.NoResultFound:
                return

    # REVISIT: This method is only called from unit tests.
    def get_ha_port_associations(self):
        session = db_api.get_reader_session()

        query = BAKERY(lambda s: s.query(
            HAIPAddressToPortAssociation))
        return query(session).all()

    # REVISIT: Move this method to the mechanism_driver or rpc module,
    # as it is above the DB level. This will also require some rework
    # of its unit tests.
    def update_ip_owner(self, ip_owner_info):
        ports_to_update = set()
        port_id = ip_owner_info.get('port')
        ipv4 = ip_owner_info.get('ip_address_v4')
        ipv6 = ip_owner_info.get('ip_address_v6')
        network_id = ip_owner_info.get('network_id')
        if not port_id or (not ipv4 and not ipv6):
            return ports_to_update
        LOG.debug("Got IP owner update: %s", ip_owner_info)
        # REVISIT: Just use SQLAlchemy session and models_v2.Port?
        port = self.plugin.get_port(n_context.get_admin_context(), port_id)
        if not port:
            LOG.debug("Ignoring update for non-existent port: %s", port_id)
            return ports_to_update
        ports_to_update.add(port_id)
        for ipa in [ipv4, ipv6]:
            if not ipa:
                continue
            try:
                # REVISIT: Why isn't this a single transaction at the
                # top-level, so that the port itself is guaranteed to
                # still exist.
                session = db_api.get_writer_session()
                with session.begin(subtransactions=True):
                    old_owner = self.get_port_for_ha_ipaddress(
                        ipa, network_id or port['network_id'], session=session)
                    self.set_port_id_for_ha_ipaddress(port_id, ipa, session)
                    if old_owner and old_owner['port_id'] != port_id:
                        self.delete_port_id_for_ha_ipaddress(
                            old_owner['port_id'], ipa, session=session)
                        ports_to_update.add(old_owner['port_id'])
            except db_exc.DBReferenceError as dbe:
                LOG.debug("Ignoring FK error for port %s: %s", port_id, dbe)
        return ports_to_update

    # VMName functions.

    def _get_vm_name(self, session, device_id, is_detailed=False):
        if is_detailed:
            query = BAKERY(lambda s: s.query(VMName))
        else:
            query = BAKERY(lambda s: s.query(VMName.vm_name))
        query += lambda q: q.filter_by(
            device_id=sa.bindparam('device_id'))
        return query(session).params(
            device_id=device_id).one_or_none()

    def _get_vm_names(self, session):
        query = BAKERY(lambda s: s.query(VMName.device_id,
                                         VMName.vm_name))
        return query(session).all()

    def _set_vm_name(self, session, device_id, vm_name):
        with session.begin(subtransactions=True):
            db_obj = self._get_vm_name(session, device_id,
                                       is_detailed=True)
            if db_obj:
                db_obj.vm_name = vm_name
            else:
                db_obj = VMName(device_id=device_id, vm_name=vm_name)
            session.add(db_obj)

    def _delete_vm_name(self, session, device_id):
        with session.begin(subtransactions=True):
            db_obj = self._get_vm_name(session, device_id,
                                       is_detailed=True)
            if db_obj:
                session.delete(db_obj)

    # VMNameUpdate functions.

    def _get_vm_name_update(self, session):
        query = BAKERY(lambda s: s.query(VMNameUpdate))
        return query(session).one_or_none()

    def _set_vm_name_update(self, session, db_obj, host_id,
                            last_incremental_update_time,
                            last_full_update_time=None):
        with session.begin(subtransactions=True):
            if db_obj:
                db_obj.host_id = host_id
                db_obj.last_incremental_update_time = (
                                    last_incremental_update_time)
                if last_full_update_time:
                    db_obj.last_full_update_time = last_full_update_time
            else:
                db_obj = VMNameUpdate(
                    purpose=VM_UPDATE_PURPOSE, host_id=host_id,
                    last_incremental_update_time=last_incremental_update_time,
                    last_full_update_time=last_full_update_time)
            session.add(db_obj)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import collections

from neutron.common import config
from neutron_lib import context as n_ctx
from oslo_config import cfg
from oslo_serialization import jsonutils

from gbpservice.neutron.services.servicechain.plugins.ncp import (
    plugin as ncp_plugin)
from gbpservice.neutron.services.servicechain.plugins.ncp import context
from gbpservice.neutron.tests.unit.db.grouppolicy import (
    test_servicechain_db as test_servicechain_db)
from gbpservice.neutron.tests.unit.db.grouppolicy import test_group_policy_db

cfg.CONF.import_opt(
    'node_drivers',
    'gbpservice.neutron.services.servicechain.plugins.ncp.config',
    group='node_composition_plugin')


class ServiceChainNCPTestPlugin(ncp_plugin.NodeCompositionPlugin):

    supported_extension_aliases = ['servicechain'] + (
        test_group_policy_db.UNSUPPORTED_REQUIRED_EXTS)
    path_prefix = "/servicechain"


SC_PLUGIN_KLASS = (ServiceChainNCPTestPlugin.__module__ + '.' +
                   ServiceChainNCPTestPlugin.__name__)


class ServiceChainPluginTestCase(test_servicechain_db.ServiceChainDbTestCase):

    def setUp(self, core_plugin=None, sc_plugin=None, gp_plugin=None):
        super(ServiceChainPluginTestCase, self).setUp(core_plugin=core_plugin,
                                                      sc_plugin=sc_plugin or
                                                      SC_PLUGIN_KLASS,
                                                      gp_plugin=gp_plugin)
        try:
            config.cfg.CONF.keystone_authtoken.username
        except config.cfg.NoSuchOptError:
            config.cfg.CONF.register_opt(
                config.cfg.StrOpt('username'),
                'keystone_authtoken')
        try:
            config.cfg.CONF.keystone_authtoken.password
        except config.cfg.NoSuchOptError:
            config.cfg.CONF.register_opt(
                config.cfg.StrOpt('password'),
                'keystone_authtoken')
        try:
            config.cfg.CONF.keystone_authtoken.project_name
        except config.cfg.NoSuchOptError:
            config.cfg.CONF.register_opt(
                config.cfg.StrOpt('project_name'),
                'keystone_authtoken')


class BaseTestGroupPolicyPluginGroupResources(
        ServiceChainPluginTestCase,
        test_servicechain_db.TestServiceChainResources):

    def test_spec_shared(self):
        # Shared spec can only point shared nodes
        node = self._create_profiled_servicechain_node(
            'LOADBALANCERV2', shared=True, shared_profile=True,
            profile_tenant_id='admin', tenant_id='admin')['servicechain_node']
        self.create_servicechain_spec(nodes=[node['id']], shared=True,
                                      expected_res_status=201)
        self.create_servicechain_spec(nodes=[node['id']], shared=False,
                                      tenant_id='admin',
                                      expected_res_status=201)

        node = self._create_profiled_servicechain_node(
            'LOADBALANCERV2', shared=False, profile_tenant_id='nonadmin',
            tenant_id='nonadmin')['servicechain_node']
        self.create_servicechain_spec(nodes=[node['id']], shared=True,
                                      expected_res_status=404)
        self.create_servicechain_spec(nodes=[node['id']], shared=True,
                                      tenant_id='nonadmin',
                                      expected_res_status=400)
        self.create_servicechain_spec(nodes=[node['id']], shared=False,
                                      tenant_id='nonadmin',
                                      expected_res_status=201)

    def test_node_shared(self):
        # Shared node can only point shared profile
        prof = self.create_service_profile(
            service_type='LOADBALANCERV2', shared=True,
            tenant_id='admin')['service_profile']
        to_update = self.create_servicechain_node(
            service_profile_id=prof['id'], shared=True,
            expected_res_status=201)['servicechain_node']
        self.create_servicechain_node(
            service_profile_id=prof['id'], shared=False, tenant_id='admin',
            expected_res_status=201)

        prof = self.create_service_profile(
            service_type='LOADBALANCERV2', shared=False,
            tenant_id='admin')['service_profile']
        self.create_servicechain_node(
            service_profile_id=prof['id'], shared=True,
            expected_res_status=404)
        self.create_servicechain_node(
            service_profile_id=prof['id'], shared=True,
            tenant_id='admin', expected_res_status=400)
        self.create_servicechain_node(
            service_profile_id=prof['id'], shared=False,
            tenant_id='admin', expected_res_status=201)

        self.create_servicechain_spec(nodes=[to_update['id']], shared=True,
                                      tenant_id='nonadmin',
                                      expected_res_status=201)

        data = {'servicechain_node': {'shared': False}}
        req = self.new_update_request('servicechain_nodes', data,
                                      to_update['id'])
        res = req.get_response(self.ext_api)
        self.assertEqual(400, res.status_int)
        res = self.deserialize(self.fmt, res)
        self.assertEqual('InvalidSharedAttributeUpdate',
                         res['NeutronError']['type'])

    def test_profile_shared(self):
        prof = self.create_service_profile(
            service_type='LOADBALANCERV2', shared=True,
            tenant_id='admin')['service_profile']
        self.create_servicechain_node(
            service_profile_id=prof['id'], shared=True,
            expected_res_status=201)

        data = {'service_profile': {'shared': False}}
        req = self.new_update_request('service_profiles', data,
                                      prof['id'])
        res = req.get_response(self.ext_api)
        self.assertEqual(400, res.status_int)
        res = self.deserialize(self.fmt, res)
        self.assertEqual('InvalidSharedAttributeUpdate',
                         res['NeutronError']['type'])

        prof = self.create_service_profile(
            service_type='LOADBALANCERV2', shared=False)['service_profile']
        self.create_servicechain_node(
            service_profile_id=prof['id'], shared=False,
            expected_res_status=201)

        data = {'service_profile': {'shared': True}}
        req = self.new_update_request('service_profiles', data,
                                      prof['id'])
        res = req.get_response(self.ext_api)
        self.assertEqual(200, res.status_int)
        res = self.deserialize(self.fmt, res)
        self.assertTrue(res['service_profile']['shared'])

    def test_node_context_profile(self):

        # Current node with profile
        plugin_context = n_ctx.get_admin_context()
        plugin_context.is_admin = plugin_context.is_advsvc = False
        plugin_context.tenant_id = self._tenant_id

        prof = self.create_service_profile(
            service_type='LOADBALANCERV2')['service_profile']
        current = self.create_servicechain_node(
            service_profile_id=prof['id'],
            expected_res_status=201)['servicechain_node']
        ctx = context.NodeDriverContext(self.plugin, plugin_context,
                                        None, None, current, 0,
                                        prof, None)

        self.assertIsNone(ctx.original_node)
        self.assertIsNone(ctx.original_profile)
        self.assertEqual(ctx.current_node, current)
        self.assertEqual(ctx.current_profile, prof)

        # Original node with profile

        prof2 = self.create_service_profile(
            service_type='LOADBALANCERV2')['service_profile']
        original = self.create_servicechain_node(
            service_profile_id=prof2['id'],
            expected_res_status=201)['servicechain_node']
        ctx = context.NodeDriverContext(
                self.plugin, plugin_context, None, None, current, 0,
                prof, None, original_service_chain_node=original,
                original_service_profile=prof2)

        self.assertEqual(ctx.original_node, original)
        self.assertEqual(ctx.original_profile, prof2)
        self.assertEqual(ctx.current_node, current)
        self.assertEqual(ctx.current_profile, prof)

    def test_node_context_no_profile(self):

        plugin_context = n_ctx.get_admin_context()
        plugin_context.is_admin = plugin_context.is_advsvc = False
        plugin_context.tenant_id = 'test_tenant'

        current = self.create_servicechain_node(
            service_type='TEST',
            expected_res_status=201)['servicechain_node']
        ctx = context.NodeDriverContext(self.plugin, plugin_context,
                                        None, None, current, 0,
                                        None, None)

        self.assertIsNone(ctx.original_node)
        self.assertIsNone(ctx.original_profile)
        self.assertEqual(ctx.current_node, current)
        self.assertIsNone(ctx.current_profile)

        original = self.create_servicechain_node(
            service_type='TEST',
            expected_res_status=201)['servicechain_node']
        ctx = context.NodeDriverContext(
                self.plugin, plugin_context, None, None, current, 0,
                None, None, original_service_chain_node=original)

        self.assertEqual(ctx.original_node, original)
        self.assertIsNone(ctx.original_profile)
        self.assertEqual(ctx.current_node, current)
        self.assertIsNone(ctx.current_profile)

    def test_spec_parameters(self):
        params_node_1 = ['p1', 'p2', 'p3']
        params_node_2 = ['p4', 'p5', 'p6']
        params_node_3 = ['p7', 'p8', 'p9']

        def params_dict(params):
            return jsonutils.dumps({'Parameters':
                                    dict((x, {}) for x in params)})

        prof = self.create_service_profile(
            service_type='LOADBALANCERV2', shared=True,
            tenant_id='admin')['service_profile']

        # Create 2 nodes with different parameters
        node1 = self.create_servicechain_node(
            service_profile_id=prof['id'], shared=True,
            config=params_dict(params_node_1),
            expected_res_status=201)['servicechain_node']
        node2 = self.create_servicechain_node(
            service_profile_id=prof['id'], shared=True,
            config=params_dict(params_node_2),
            expected_res_status=201)['servicechain_node']

        # Create SC spec with the nodes assigned
        spec = self.create_servicechain_spec(
            nodes=[node1['id'], node2['id']], shared=True,
            expected_res_status=201)['servicechain_spec']

        # Verify param names correspondence
        self.assertEqual(
            collections.Counter(params_node_1 + params_node_2),
            collections.Counter(ast.literal_eval(spec['config_param_names'])))

        # Update the spec removing one node
        self.update_servicechain_spec(spec['id'], nodes=[node1['id']],
                                      expected_res_status=200)

        spec = self.show_servicechain_spec(spec['id'])['servicechain_spec']
        # Verify param names correspondence
        self.assertEqual(
            collections.Counter(params_node_1),
            collections.Counter(ast.literal_eval(spec['config_param_names'])))

        # Update the spec without modifying the node list
        self.update_servicechain_spec(spec['id'],
                                      name='new_name',
                                      expected_res_status=200)

        spec = self.show_servicechain_spec(spec['id'])['servicechain_spec']
        # Verify param names correspondence
        self.assertEqual(
            collections.Counter(params_node_1),
            collections.Counter(ast.literal_eval(spec['config_param_names'])))

        # Update a node with new config params
        self.update_servicechain_node(node1['id'],
                                      config=params_dict(params_node_3),
                                      expected_res_status=200)

        spec = self.show_servicechain_spec(spec['id'])['servicechain_spec']
        # Verify param names correspondence
        self.assertEqual(
            collections.Counter(params_node_3),
            collections.Counter(ast.literal_eval(spec['config_param_names'])))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import converters as conv
from neutron_lib.api import extensions

from gbpservice._i18n import _
from gbpservice.neutron.extensions import cisco_apic
from gbpservice.neutron.extensions import group_policy as gp

ALIAS = 'cisco-apic-gbp'

FORWARD_FILTER_ENTRIES = 'Forward-FilterEntries'
REVERSE_FILTER_ENTRIES = 'Reverse-FilterEntries'
CONTRACT = 'Contract'
CONTRACT_SUBJECT = 'ContractSubject'

EXTENDED_ATTRIBUTES_2_0 = {
    gp.POLICY_TARGET_GROUPS: {
        cisco_apic.DIST_NAMES: {
            'allow_post': False, 'allow_put': False, 'is_visible': True},
        'intra_ptg_allow': {
            'allow_post': True, 'allow_put': True, 'default': True,
            'convert_to': conv.convert_to_boolean, 'is_visible': True},
        'is_auto_ptg': {
            'allow_post': False, 'allow_put': False,
            'convert_to': conv.convert_to_boolean, 'is_visible': True,
            'enforce_policy': True},
    },
    gp.POLICY_RULES: {
        cisco_apic.DIST_NAMES: {
            'allow_post': False, 'allow_put': False, 'is_visible': True},
    },
    gp.POLICY_RULE_SETS: {
        cisco_apic.DIST_NAMES: {
            'allow_post': False, 'allow_put': False, 'is_visible': True},
    },
    gp.APPLICATION_POLICY_GROUPS: {
        cisco_apic.DIST_NAMES: {
            'allow_post': False, 'allow_put': False, 'is_visible': True},
    },
}


class Cisco_apic_gbp(extensions.ExtensionDescriptor):

    @classmethod
    def get_name(cls):
        return "Cisco APIC GBP"

    @classmethod
    def get_alias(cls):
        return ALIAS

    @classmethod
    def get_description(cls):
        return _("Extension exposing mapping of GBP resources to Cisco "
                 "APIC constructs")

    @classmethod
    def get_updated(cls):
        return "2016-07-11T10:00:00-00:00"

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}



# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gbpclient.v2_0 import client as gbp_client
from keystoneclient import auth as ksc_auth
from keystoneclient import session as ksc_session
from keystoneclient.v3 import client as ksc_client
from oslo_config import cfg
from oslo_log import log as logging


LOG = logging.getLogger(__name__)

# REVISIT(rkukura): We use keystone to get the name of the keystone
# project owning each neutron resource, which by default, requires
# admin. If we keep this, we should probably move it to a separate
# config module. But we should also investigate whether admin is even
# needed, or if neutron's credentials could somehow be used.
AUTH_GROUP = 'apic_aim_auth'
ksc_session.Session.register_conf_options(cfg.CONF, AUTH_GROUP)
ksc_auth.register_conf_options(cfg.CONF, AUTH_GROUP)


class ProjectDetailsCache(object):
    """Cache of Keystone project ID to project details mappings."""

    def __init__(self):
        self.project_details = {}
        self.keystone = None
        self.gbp = None
        self.enable_neutronclient_internal_ep_interface = (
            cfg.CONF.ml2_apic_aim.enable_neutronclient_internal_ep_interface)

    def _get_keystone_client(self):
        # REVISIT: It seems load_from_conf_options() and
        # keystoneclient auth plugins have been deprecated, and we
        # should use keystoneauth instead.
        LOG.debug("Getting keystone client")
        auth = ksc_auth.load_from_conf_options(cfg.CONF, AUTH_GROUP)
        LOG.debug("Got auth: %s", auth)
        if not auth:
            LOG.warning('No auth_plugin configured in %s',
                        AUTH_GROUP)
        session = ksc_session.Session.load_from_conf_options(
            cfg.CONF, AUTH_GROUP, auth=auth)
        LOG.debug("Got session: %s", session)
        self.keystone = ksc_client.Client(session=session)
        LOG.debug("Got keystone client: %s", self.keystone)
        endpoint_type = 'publicURL'
        if self.enable_neutronclient_internal_ep_interface:
            endpoint_type = 'internalURL'
        self.gbp = gbp_client.Client(session=session,
                                     endpoint_type=endpoint_type)
        LOG.debug("Got gbp client: %s", self.gbp)

    def ensure_project(self, project_id):
        """Ensure cache contains mapping for project.

        :param project_id: ID of the project

        Ensure that the cache contains a mapping for the project
        identified by project_id. If it is not, Keystone will be
        queried for the current list of projects, and any new mappings
        will be added to the cache. This method should never be called
        inside a transaction with a project_id not already in the
        cache.
        """
        if project_id and project_id not in self.project_details:
            self.load_projects()

    def load_projects(self):
        if self.keystone is None:
            self._get_keystone_client()
        LOG.debug("Calling project API")
        projects = self.keystone.projects.list()
        LOG.debug("Received projects: %s", projects)
        for project in projects:
            self.project_details[project.id] = (project.name,
                project.description)

    def get_project_details(self, project_id):
        """Get name and descr of project from cache.

        :param project_id: ID of the project

        If the cache contains project_id, a tuple with
        project name and description is returned
        else a tuple (None,None) is returned
        """
        if self.project_details.get(project_id):
            return self.project_details[project_id]
        return ('', '')

    def update_project_details(self, project_id):
        """Update project name and description from keystone.

        :param project_id: ID of the project

        Get the name and description of the project identified by
        project_id from the keystone. If the value in cache doesn't
        match values in keystone, update the cache and return 1,
        to indicate that cache has been updated
        """
        if self.keystone is None:
            self._get_keystone_client()
        if self.keystone:
            project = self.keystone.projects.get(project_id)
            if project:
                prj_details = self.get_project_details(project_id)
                if (prj_details[0] != project.name or
                    prj_details[1] != project.description):
                    self.project_details[project_id] = (
                        project.name, project.description)
                    LOG.debug("Project updated %s ",
                              str(self.project_details[project_id]))
                    return self.project_details[project_id]
        return None

    def purge_gbp(self, project_id):
        class TempArg(object):
            pass

        if self.gbp is None:
            self._get_keystone_client()
        if self.gbp:
            LOG.debug("Calling gbp purge() API")
            temp_arg = TempArg()
            temp_arg.tenant = project_id
            self.gbp.purge(temp_arg)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import commands
import logging
import sys

from libs import config_libs
from libs import utils_libs
from libs import verify_libs


def main():

    # Run the Testcases:
    test = test_gbp_l2p_func()
    if test.test_gbp_l2p_func_1() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_L2P_FUNC_1')
    if test.test_gbp_l2p_func_2() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_L2P_FUNC_2')
    test.cleanup()
    utils_libs.report_results('test_gbp_l2p_func', 'test_results.txt')
    sys.exit(1)


class test_gbp_l2p_func(object):

    # Initialize logging
    logging.basicConfig(
        format='%(asctime)s [%(levelname)s] %(name)s - %(message)s',
        level=logging.WARNING)
    _log = logging.getLogger(__name__)
    cmd = 'rm /tmp/test_gbp_l2p_func.log'
    commands.getoutput(cmd)
    hdlr = logging.FileHandler('/tmp/test_gbp_l2p_func.log')
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    hdlr.setFormatter(formatter)
    _log.addHandler(hdlr)
    _log.setLevel(logging.INFO)
    _log.setLevel(logging.DEBUG)

    def __init__(self):
        """
        Init def
        """
        self._log.info("\n## START OF GBP L3_POLICY FUNCTIONALITY TESTSUITE\n")
        self.gbpcfg = config_libs.Gbp_Config()
        self.gbpverify = verify_libs.Gbp_Verify()
        self.l3p_name = 'demo_l3p'
        self.l2p_name = 'demo_l2p'

    def cleanup(self, tc_name=''):
        if tc_name != '':
            self._log.info('%s: FAILED' % (tc_name))
        for obj in ['group', 'l2p', 'l3p']:
            self.gbpcfg.gbp_del_all_anyobj(obj)

    def test_gbp_l2p_func_1(self):

        self._log.info(
            "\n############################################################\n"
            "TESTCASE_GBP_L2P_FUNC_1: TO CREATE/VERIFY/DELETE/VERIFY "
            "a L2POLICY with DEFAULT ATTRIB VALUE\n"
            "TEST_STEPS::\n"
            "Create L2 Policy Object with default attributes\n"
            "Verify the attributes & value, show & list cmds\n"
            "Verify the implicit neutron objects and gbp L3P\n"
            "Delete L2 Policy\n"
            "Verify that L2P,L3P and implicit neutron objects has got "
            "deleted, show & list cmds\n"
            "##############################################################\n")

        # Testcase work-flow starts
        # Create and Verify L2Policy with default attrs(L3Policy & implicit
        # Neutron net obj)
        self._log.info(
            '\n## Step 1: Create L2Policy with default attrib vals##\n')
        uuids = self.gbpcfg.gbp_policy_cfg_all(1, 'l2p', self.l2p_name)
        if uuids == 0:
            self._log.info("\n## Step 1: Create L2Policy == Failed")
            return 0
        elif len(uuids) < 2:
            self._log.info(
                "\n## Step 1:Create L2Policy Failed due to unexpected "
                "tuple length\n")
            return 0
        else:
            l2p_uuid, def_l3p_uuid = uuids[0], uuids[1]
        self._log.info(
            "\n## Step 2: Verify L2Policy, default L3Policy, and Implicit "
            "Neutron objs")
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                0, 'l2p', l2p_uuid, self.l2p_name) == 0:
            self._log.info(
                "\n## Step 2A: Verify L2Policy using -list option == Failed")
            return 0
        net_uuid = self.gbpverify.gbp_l2l3ntk_pol_ver_all(
            1,
            'l2p',
            l2p_uuid,
            ret='default',
            id=l2p_uuid,
            name=self.l2p_name,
            l3_policy_id=def_l3p_uuid)[1]
        if net_uuid != 0 and isinstance(net_uuid, str):
            if self.gbpverify.neut_ver_all(
                    'net',
                    net_uuid,
                    name='l2p_%s' %
                    (self.l2p_name),
                    admin_state_up='True',
                    status='ACTIVE',
                    shared='False') == 0:
                self._log.info(
                    "\n## Step 2C: Verify implicit neutron network "
                    "object == Failed")
                return 0
        else:
            self._log.info(
                "\n## Step 2B: Verify L2Policy using -show option == Failed")
            return 0
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                1,
                'l3p',
                def_l3p_uuid,
                name='default',
                id=def_l3p_uuid,
                l2_policies=l2p_uuid) == 0:
            self._log.info("\n## Step 2D: Verify default L3Policy == Failed")
            return 0

        # Delete the L2Policy and verify L2P,def l3P and Neutron obj are
        # deleted too
        self._log.info(
            '\n## Step 3: Delete L2Policy and Verify L2P, default L3P and '
            'Implicit Neutron deleted ##\n')
        if self.gbpcfg.gbp_policy_cfg_all(0, 'l2p', self.l2p_name) == 0:
            self._log.info("\n## Step 3: Delete L2Policy == Failed")
            return 0
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(1, 'l2p', l2p_uuid) != 0:
            self._log.info(
                "\n## Step 3A: Verify L2Policy is Deleted using -show "
                "option == Failed")
            return 0
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(1, 'l3p', def_l3p_uuid) != 0:
            self._log.info(
                "\n## Step 3B: Verify default L3Policy is Deleted == Failed")
            return 0
        if self.gbpverify.neut_ver_all('net', net_uuid) != 0:
            self._log.info(
                "\n## Step 3C: Verify Implicit Neutron Network Obj is "
                "Deleted == Failed")
            return 0
        self._log.info("\n## TESTCASE_GBP_L2P_FUNC_1: PASSED")
        return 1

    def test_gbp_l2p_func_2(self):

        self._log.info(
            "\n############################################################\n"
            "TESTCASE_GBP_L2P_FUNC_2: TO CREATE/UPDATE/DELETE/VERIFY a "
            "L2POLICY with MULTIPLE PTGs\n"
            "TEST_STEPS::\n"
            "Create L2Policy Object with non-default params\n"
            "Verify the attributes & value, show & list cmds\n"
            "Create Multiple(n=10) PTGs using the above L2P\n"
            "Verify the PTGs and L2P are reflecting in each other in the DB\n"
            "Delete all PTG, L2P\n"
            "Verify that all Implicit Neutron Net/Subnets, default L3P has "
            "got deleted along with PTGs & L2P\n"
            "##############################################################\n")

        # Testcase work-flow starts
        # Create and Verify L2Policy with default attrs(L3Policy & implicit
        # Neutron net obj)
        self._log.info(
            '\n## Step 1: Create L2Policy with default attrib vals##\n')
        uuids = self.gbpcfg.gbp_policy_cfg_all(1, 'l2p', self.l2p_name)
        if uuids == 0:
            self._log.info("\n## Step 1: Create L2Policy == Failed")
            return 0
        elif len(uuids) < 2:
            self._log.info(
                "\n## Step 1:Create L2Policy Failed due to unexpected "
                "tuple length\n")
            return 0
        else:
            l2p_uuid, def_l3p_uuid = uuids[0], uuids[1]
        self._log.info(
            "\n## Step 2: Verify L2Policy, default L3Policy, and Implicit "
            "Neutron objs")
        net_uuid = self.gbpverify.gbp_l2l3ntk_pol_ver_all(
            1,
            'l2p',
            l2p_uuid,
            ret='default',
            id=l2p_uuid,
            name=self.l2p_name,
            l3_policy_id=def_l3p_uuid)[1]
        if net_uuid != 0 and isinstance(net_uuid, str):
            if self.gbpverify.neut_ver_all(
                    'net',
                    net_uuid,
                    name='l2p_%s' %
                    (self.l2p_name),
                    admin_state_up='True',
                    status='ACTIVE',
                    shared='False') == 0:
                self._log.info(
                    "\n## Step 2A: Verify implicit neutron network "
                    "object == Failed")
                return 0
        else:
            self._log.info(
                "\n## Step 2B: Verify L2Policy using -show option == Failed")
            return 0
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                1,
                'l3p',
                def_l3p_uuid,
                name='default',
                id=def_l3p_uuid,
                l2_policies=l2p_uuid) == 0:
            self._log.info("\n## Step 2C: Verify default L3Policy == Failed")
            return 0
        # Create Multiple PTGs and associate the above L2P
        self._log.info(
            "\n## Step 3: Create Multiple(n=10) PTGs using the above L2P\n")
        n, i = 11, 1
        ptg_list = []
        subnet_list = []
        while i < n:
            ptg_name = 'demo_ptg_%s' % (i)
            uuids = self.gbpcfg.gbp_policy_cfg_all(
                1, 'group', ptg_name, l2_policy=l2p_uuid)
            if uuids != 0:
                ptg_list.append(uuids[0])
                subnet_list.append(uuids[2])
            else:
                self._log.info("\n## Step 3: Create Target-Group == Failed")
                return 0
            if self.gbpverify.gbp_policy_verify_all(
                1, 'group', ptg_name, id=ptg_list[
                    i - 1], shared='False', l2_policy_id=l2p_uuid) == 0:
                self._log.info(
                    "\n## Step 3A: Verify Policy Target-Group "
                    "using L2P == Failed")
                return 0
            if self.gbpverify.neut_ver_all(
                'subnet', subnet_list[
                    i - 1], network_id=net_uuid) == 0:
                self._log.info(
                    "\n## Step 3B: Verify Implicit Neutron Subnet == Failed")
                return 0
            i += 1
        if self.gbpverify.gbp_obj_ver_attr_all_values(
                'l2p', l2p_uuid, 'policy_target_groups', ptg_list) == 0:
            self._log.info(
                "\n## Step 3C: Verifying multiple PTGs mapped to "
                "L2P == Failed \n")
            return 0
        # Delete all PTGs and L2P
        self._log.info("\n## Step 4: Delete all PTGs and L2P\n")
        for ptgid in ptg_list:
            if self.gbpcfg.gbp_policy_cfg_all(0, 'group', ptgid) == 0:
                self._log.info("\n## Step 4: Delete Target-Group == Failed")
                return 0
        if self.gbpcfg.gbp_policy_cfg_all(0, 'l2p', l2p_uuid) == 0:
            self._log.info("\n## Step 4A: Delete L2Policy == Failed")
            return 0
        # Verify that all Implicit Neutron Objs, Default L3p , PTGs and L2P are
        # deleted
        self._log.info(
            "\n## Step 5: Verify that all Implicit Neutron Objs, Default L3p, "
            "PTGs and L2P are deleted\n")
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(1, 'l3p', 'default') != 0:
            self._log.info(
                "\n## Step 5: Verify that default L3P has got auto-deleted "
                "== Failed\n")
            return 0
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(1, 'l2p', l2p_uuid) != 0:
            self._log.info(
                "\n## Step 5A: Verify that L2P got deleted == Failed\n")
            return 0
        for subnet in subnet_list:
            if self.gbpverify.neut_ver_all('subnet', subnet) != 0:
                self._log.info(
                    "\n## Step 5B: Verify that Implicit Neutron Subnet got "
                    "deleted == Failed\n")
                return 0
        for ptgid in ptg_list:
            if self.gbpverify.gbp_policy_verify_all(1, 'group', ptgid) != 0:
                self._log.info(
                    "\n## Step 3B: Verify Target-Group is Deleted using "
                    "-show option == Failed")
                return 0
        self._log.info("\n## TESTCASE_GBP_L2P_FUNC_2: PASSED")
        return 1


if __name__ == '__main__':
    main()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""allowed_vm_names

Revision ID: 5629167be1d1
Revises: 092e4b1aeb0a
Create Date: 2016-10-11 14:14:06.648609

"""

# revision identifiers, used by Alembic.
revision = '5629167be1d1'
down_revision = '092e4b1aeb0a'

from alembic import op
import sqlalchemy as sa


def upgrade():

    op.create_table(
        'gp_apic_mapping_allowed_vm_names',
        sa.Column('l3_policy_id', sa.String(length=36), nullable=False),
        sa.Column('allowed_vm_name', sa.String(length=255),
                  nullable=False),
        sa.ForeignKeyConstraint(
            ['l3_policy_id'], ['gp_l3_policies.id'],
            name='gp_apic_mapping_allowed_vm_name_fk_l3pid',
            ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('l3_policy_id', 'allowed_vm_name')
    )


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import extensions

from gbpservice._i18n import _
from gbpservice.neutron.extensions import group_policy as gp


CISCO_APIC_GBP_REUSE_BD_EXT = 'cisco_apic_gbp_reuse_bd'

EXTENDED_ATTRIBUTES_2_0 = {
    gp.L2_POLICIES: {
        'reuse_bd': {
            'allow_post': True, 'allow_put': False, 'default': None,
            'validate': {'type:uuid_or_none': None},
            'is_visible': True},
    },
}


class Apic_reuse_bd(extensions.ExtensionDescriptor):

    @classmethod
    def get_name(cls):
        return "APIC GBP Reuse BD Extension"

    @classmethod
    def get_alias(cls):
        return CISCO_APIC_GBP_REUSE_BD_EXT

    @classmethod
    def get_description(cls):
        return _("This extension enables creating L2 policy objects that "
                 "use the same BridgeDomain on APIC")

    @classmethod
    def get_updated(cls):
        return "2016-11-11T04:20:00-00:00"

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.plugins.ml2.drivers.openvswitch.mech_driver import (
        mech_openvswitch as base)
from neutron_lib.api.definitions import portbindings
from neutron_lib.plugins.ml2 import api

from gbpservice.neutron.services.servicechain.plugins.ncp import plumber_base


class TrafficStitchingMechanismGBPDriver(base.OpenvswitchMechanismDriver):
    """Traffic Stitching Mechanism Driver for GBP.

    This driver makes sure that service targets are bound with port_filter and
    hybrid_mode set to false. This should disable port security and anti
    spoofing rules for these special ports.
    """

    def try_to_bind_segment_for_agent(self, context, segment, agent):
        vif_details = self.vif_details
        if self.check_segment_for_agent(segment, agent):
            if context.current['name'].startswith(
                    'pt_' + plumber_base.SERVICE_TARGET_NAME_PREFIX):
                vif_details = {portbindings.CAP_PORT_FILTER: False,
                               portbindings.OVS_HYBRID_PLUG: False}
            context.set_binding(
                segment[api.ID], self.vif_type, vif_details)
            return True
        else:
            return False



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import commands
import logging
import sys

from libs import config_libs
from libs import utils_libs
from libs import verify_libs


def main():

    # Run the Testcase:
    test = test_gbp_pc_neg()
    if test.test_gbp_pc_neg_1() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PC_NEG_1')
    if test.test_gbp_pc_neg_2() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PC_NEG_2')
    if test.test_gbp_pc_neg_3() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PC_NEG_3')
    if test.test_gbp_pc_neg_4() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PC_NEG_4')
    if test.test_gbp_pc_neg_5() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PC_NEG_5')
    if test.test_gbp_pc_neg_6() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PC_NEG_6')
    test.cleanup()
    utils_libs.report_results('test_gbp_pc_neg', 'test_results.txt')
    sys.exit(1)


class test_gbp_pc_neg(object):

    # Initialize logging
    logging.basicConfig(
        format='%(asctime)s [%(levelname)s] %(name)s - %(message)s',
        level=logging.WARNING)
    _log = logging.getLogger(__name__)
    cmd = 'rm /tmp/test_gbp_pc_neg.log'
    commands.getoutput(cmd)
    hdlr = logging.FileHandler('/tmp/test_gbp_pc_neg.log')
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    hdlr.setFormatter(formatter)
    _log.addHandler(hdlr)
    _log.setLevel(logging.INFO)
    _log.setLevel(logging.DEBUG)

    def __init__(self):
        """
        Init def
        """
        self._log.info(
            "\n## START OF GBP POLICY_CLASSIFIER NEGATIVE TESTSUITE\n")
        self.gbpcfg = config_libs.Gbp_Config()
        self.gbpverify = verify_libs.Gbp_Verify()
        self.cls_name = 'demo_pc'

    def cleanup(self, tc_name=''):
        if tc_name != '':
            self._log.info('%s FAILED' % (tc_name))
        for obj in ['classifier']:
            self.gbpcfg.gbp_del_all_anyobj(obj)

    def test_gbp_pc_neg_1(self):

        self._log.info(
            "\n#######################################################\n"
            "TESTCASE_GBP_PC_NEG_1: CREATE/VERIFY a "
            "POLICY CLASSIFIER with INVALID PROTO \n"
            "TEST_STEP::\n"
            "Create Policy Classifier with Invalid "
            "Proto(any proto other than tcp,udp,icmp)\n"
            "Verify that the create fails and rollbacks\n"
            "#######################################################\n")

        # Testcase work-flow starts
        self._log.info(
            "\n## Step 1: Create Classifier with invalid protocol##\n")
        if self.gbpcfg.gbp_policy_cfg_all(
                1, 'classifier', self.cls_name, protocol='http') != 0:
            self._log.info(
                "\n## Step 1: Create Classifier with Invalid Protocol "
                "did NOT Fail")
            return 0
        self._log.info("\n## Step 1A: Verify classifier has been rolled back")
        if self.gbpverify.gbp_classif_verify(1, self.cls_name) != 0:
            self._log.info("\n## Step 1A: Classifier did NOT roll back")
            return 0
        self._log.info("\n## TESTCASE_GBP_PC_NEG_1: PASSED")
        return 1

    def test_gbp_pc_neg_2(self):

        self._log.info(
            "\n#######################################################\n"
            "TESTCASE_GBP_PC_NEG_2: CREATE/VERIFY a POLICY CLASSIFIER with "
            "INVALID PORT-RANGE \n"
            "TEST_STEP::\n"
            "Create Policy Classifier with Valid Proto BUT Invalid "
            "Port-range\n"
            "Verify that the create fails and rollbacks\n"
            "#######################################################\n")

        # Testcase work-flow starts
        self._log.info(
            "\n## Step 1: Create Classifier with invalid protocol##\n")
        if self.gbpcfg.gbp_policy_cfg_all(
                1,
                'classifier',
                self.cls_name,
                protocol='tcp',
                port_range='80:50') != 0:
            self._log.info(
                "\n## Step 1: Create Classifier with Invalid Port-Range "
                "did NOT Fail")
            return 0
        self._log.info("\n## Step 1A: Verify classifier has been rolled back")
        if self.gbpverify.gbp_classif_verify(1, self.cls_name) != 0:
            self._log.info("\n## Step 1A: Classifier did NOT roll back")
            return 0
        self._log.info("\n## TESTCASE_GBP_PC_NEG_2: PASSED")
        return 1

    def test_gbp_pc_neg_3(self):

        self._log.info(
            "\n#######################################################\n"
            "TESTCASE_GBP_PC_NEG_3: DELETE NON-EXISTENT/INVALID "
            "POLICY CLASSIFICER\n"
            "TEST_STEP::\n"
            "Delete unknown/invalid policy-classifier\n"
            "#######################################################\n")

        self._log.info("\n## Step 1: Delete non-existent Classifier  ##")
        if self.gbpcfg.gbp_policy_cfg_all(0, 'classifier', self.cls_name) != 0:
            self._log.info(
                "\n## Step 1: Delete Non-existent policy classifier "
                "did NOT Fail")
            return 0
        self._log.info("\n## TESTCASE_GBP_PC_NEG_3: PASSED")
        return 1

    def test_gbp_pc_neg_4(self):

        self._log.info(
            "\n#######################################################\n"
            "TESTCASE_GBP_PC_NEG_4: CREATE/VERIFY POLICY CLASSIFIER "
            "with INVALID DIRECTION \n"
            "TEST_STEP::\n"
            "Create Policy Classifier using Invalid Direction\n"
            "Verify that the create fails and rollbacks\n"
            "######################################################\n")

        # Testcase work-flow starts
        self._log.info(
            "\n## Step 1: Create Classifier with Invalid Direction##\n")
        if self.gbpcfg.gbp_policy_cfg_all(
                1,
                'classifier',
                self.cls_name,
                direction='redirect') != 0:
            self._log.info(
                "\n## Step 1: Create Classifier with Invalid Direction "
                "did NOT Fail")
            return 0
        self._log.info("\n## Step 1A: Verify classifier has been rolled back")
        if self.gbpverify.gbp_classif_verify(1, self.cls_name) != 0:
            self._log.info("\n## Step 1A: Classifier did NOT roll back")
            return 0
        self._log.info("\n## TESTCASE_GBP_PC_NEG_4: PASSED")
        return 1

    def test_gbp_pc_neg_5(self):

        self._log.info(
            "\n#######################################################\n"
            "TESTCASE_GBP_PC_NEG_5: UPDATE/VERIFY/ POLICY CLASSIFIER "
            "with INVALID PROTOCOL \n"
            "TEST_STEP::\n"
            "Create Policy Classifier using non-default protocol\n"
            "Update Policy Classifier with Invalid Protocol\n"
            "Verify that the update fails and rollbacks to original values\n"
            "######################################################\n")

        # Testcase work-flow starts
        self._log.info(
            "\n## Step 1: Create Classifier with non-default protocol ##\n")
        self.cls_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'classifier', self.cls_name, protocol='tcp')
        if self.cls_uuid != 0:
            self._log.info(
                "\n## Step 1: Create Classifier Passed, UUID == %s\n" %
                (self.cls_uuid))
        else:
            self._log.info("\n## Step 1: Create Classifier == Failed")
            return 0
        self._log.info(
            "\n## Step 2: Update Policy Classifier with Invalid Protocol##\n")
        if self.gbpcfg.gbp_policy_cfg_all(
                2,
                'classifier',
                self.cls_uuid,
                name='grppol_pc',
                protocol='http') != 0:
            self._log.info(
                "\n## Step 2: Updating Policy Classifier's with "
                "Invalid Protocol did NOT Fail ")
            return 0
        if self.gbpverify.gbp_classif_verify(
                1, self.cls_name, id=self.cls_uuid, protocol='tcp') == 0:
            self._log.info(
                "\n## Step 2A: Verify Policy Classifier did NOT roll back")
            return 0
        self._log.info("\n## TESTCASE_GBP_PC_NEG_5: PASSED")
        return 1

    def test_gbp_pc_neg_6(self):

        self._log.info(
            "\n######################################################\n"
            "TESTCASE_GBP_PC_NEG_6: UPDATE/VERIFY/ POLICY CLASSIFIER "
            "with INVALID PORT-RANGE \n"
            "TEST_STEP::\n"
            "Update Policy Classifier with Invalid Port-Range\n"
            "Verify that the update fails and rollbacks to original values\n"
            "#####################################################\n")

        # Testcase work-flow starts
        self._log.info(
            "\n## Step 2: Update Policy Classifier with Invalid "
            "Port-Range##\n")
        if self.gbpcfg.gbp_policy_cfg_all(
                2,
                'classifier',
                self.cls_uuid,
                name='grppol_pc',
                port_range='4000:80') != 0:
            self._log.info(
                "\n## Step 2: Updating Policy Classifier's with "
                "Invalid Port-Range did NOT Fail ")
            return 0
        if self.gbpverify.gbp_classif_verify(
                1,
                self.cls_name,
                id=self.cls_uuid,
                protocol='tcp',
                port_range='4000:80') != 0:
            self._log.info(
                "\n## Step 2A: Verify Policy Classifier did NOT roll back")
            return 0
        self._log.info("\n## TESTCASE_GBP_PC_NEG_6: PASSED")
        return 1


if __name__ == '__main__':
    main()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import eventlet
from eventlet import greenpool
from eventlet import greenthread

from gbpservice.nfp.core import log as nfp_logging

LOG = nfp_logging.getLogger(__name__)


def _thread_done(gt, *args, **kwargs):
    kwargs['pool'].thread_done(kwargs['thread'])


"""Descriptor class for green thread """


class Thread(object):

    def __init__(self, thread, pool):
        self.thread = thread
        self.thread.link(_thread_done, pool=pool, thread=self)

    def stop(self):
        self.thread.kill()

    def wait(self):
        return self.thread.wait()

    def link(self, func, *args, **kwargs):
        self.thread.link(func, *args, **kwargs)

    def identify(self):
        return "(%d -> %s)" % (os.getpid(), 'Thread')


"""Abstract class to manage green threads """


class ThreadPool(object):

    def __init__(self, thread_pool_size=10):
        self.pool = greenpool.GreenPool(thread_pool_size)
        self.threads = []

    def dispatch(self, callback, *args, **kwargs):
        """Invokes the specified function in one of the thread """
        gt = self.pool.spawn(callback, *args, **kwargs)
        th = Thread(gt, self)
        self.threads.append(th)
        return th

    def thread_done(self, thread):
        """Invoked when thread is complete, remove it from cache """
        self.threads.remove(thread)

    def stop(self):
        """To stop the thread """
        current = greenthread.getcurrent()

        # Make a copy
        for x in self.threads[:]:
            if x is current:
                # Skipping the current thread
                continue
            try:
                x.stop()
            except Exception as ex:
                message = "Exception - %s" % (ex)
                LOG.exception(message)

    def wait(self):
        """Wait for the thread """
        current = greenthread.getcurrent()

        # Make a copy
        for x in self.threads[:]:
            if x is current:
                continue
            try:
                x.wait()
            except eventlet.greenlet.GreenletExit:
                pass
            except Exception as ex:
                message = "Unexpected exception - %r" % (ex)
                LOG.error(message)



# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.api import extensions
from neutron.common import utils as n_utils
from neutron.db import dns_db
from neutron.db import extraroute_db
from neutron.db import l3_gwmode_db
from neutron.db.models import l3 as l3_db
from neutron.quota import resource_registry
from neutron_lib.api.definitions import l3 as l3_def
from neutron_lib.api.definitions import portbindings
from neutron_lib.db import model_query
from neutron_lib.db import resource_extend
from neutron_lib.db import utils as db_utils
from neutron_lib import exceptions
from neutron_lib.plugins import constants
from neutron_lib.plugins import directory
from oslo_log import log as logging
from oslo_utils import excutils

from gbpservice._i18n import _
from gbpservice.neutron.db import api as db_api
from gbpservice.neutron import extensions as extensions_pkg
from gbpservice.neutron.extensions import cisco_apic_l3 as l3_ext
from gbpservice.neutron.plugins.ml2plus import driver_api as api_plus
from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import (
    extension_db as extn_db)
from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import (
    mechanism_driver as md)


LOG = logging.getLogger(__name__)


@resource_extend.has_resource_extenders
class ApicL3Plugin(extraroute_db.ExtraRoute_db_mixin,
                   l3_gwmode_db.L3_NAT_db_mixin,
                   extn_db.ExtensionDbMixin,
                   dns_db.DNSDbMixin):

    supported_extension_aliases = ["router", "ext-gw-mode", "extraroute",
                                   "cisco-apic-l3", "dns-integration"]

    @staticmethod
    def get_plugin_type():
        return constants.L3

    @staticmethod
    def get_plugin_description():
        return _("L3 Router Service Plugin using the APIC via AIM")

    @resource_registry.tracked_resources(router=l3_db.Router,
                                         floatingip=l3_db.FloatingIP)
    def __init__(self):
        LOG.info("APIC AIM L3 Plugin __init__")
        extensions.append_api_extensions_path(extensions_pkg.__path__)
        self._mechanism_driver = None
        super(ApicL3Plugin, self).__init__()

    @property
    def _md(self):
        if not self._mechanism_driver:
            # REVISIT(rkukura): It might be safer to search the MDs by
            # class rather than index by name, or to use a class
            # variable to find the instance.
            mech_mgr = self._core_plugin.mechanism_manager
            self._mechanism_driver = mech_mgr.mech_drivers['apic_aim'].obj
        return self._mechanism_driver

    @staticmethod
    @resource_extend.extends([l3_def.ROUTERS])
    def _extend_router_dict_apic(router_res, router_db):
        if router_res.get(api_plus.BULK_EXTENDED):
            return

        LOG.debug("APIC AIM L3 Plugin extending router dict: %s", router_res)
        plugin = directory.get_plugin(constants.L3)
        session = db_api.get_session_from_obj(router_db)
        if not session:
            session = db_api.get_writer_session()
        try:
            plugin._md.extend_router_dict(session, router_db, router_res)
            plugin._include_router_extn_attr(session, router_res)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception("APIC AIM extend_router_dict failed")

    @staticmethod
    @resource_extend.extends([l3_def.ROUTERS + '_BULK'])
    def _extend_router_dict_bulk_apic(routers, _):
        LOG.debug("APIC AIM L3 Plugin bulk extending router dict: %s",
                  routers)
        if not routers:
            return
        router_db = routers[0]
        plugin = directory.get_plugin(constants.L3)
        session = db_api.get_session_from_obj(router_db)
        if not session:
            session = db_api.get_writer_session()
        try:
            plugin._md.extend_router_dict_bulk(session, routers)
            plugin._include_router_extn_attr_bulk(session, routers)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception("APIC AIM _extend_router_dict_bulk failed")

    def _make_routers_dict(self, routers, fields=None):
        results = []
        for router in routers:
            res = self._make_router_dict(router, fields,
                                         process_extensions=False)
            res[api_plus.BULK_EXTENDED] = True
            resource_extend.apply_funcs(l3_def.ROUTERS, res, router)
            res.pop(api_plus.BULK_EXTENDED, None)
            results.append(res)

        resource_extend.apply_funcs(l3_def.ROUTERS + '_BULK',
                                    results, None)
        return results

    # Overwrite the upstream implementation to take advantage
    # of the bulk extension support.
    @db_api.retry_if_session_inactive()
    def get_routers(self, context, filters=None, fields=None,
                    sorts=None, limit=None, marker=None,
                    page_reverse=False):
        marker_obj = db_utils.get_marker_obj(self, context, 'router',
                                             limit, marker)
        routers_db = model_query.get_collection(context, l3_db.Router,
                                                dict_func=None,
                                                filters=filters,
                                                fields=fields,
                                                sorts=sorts,
                                                limit=limit,
                                                marker_obj=marker_obj,
                                                page_reverse=page_reverse)
        return self._make_routers_dict(routers_db, fields)

    # REVISIT: Eliminate this wrapper?
    @n_utils.transaction_guard
    def create_router(self, context, router):
        LOG.debug("APIC AIM L3 Plugin creating router: %s", router)
        # REVISIT: Do this in MD by registering for
        # ROUTER.BEFORE_CREATE event.
        self._md.ensure_tenant(context, router['router']['tenant_id'])
        return super(ApicL3Plugin, self).create_router(context, router)

    # REVISIT: Eliminate this wrapper?
    @n_utils.transaction_guard
    def update_router(self, context, id, router):
        LOG.debug("APIC AIM L3 Plugin updating router %(id)s with: %(router)s",
                  {'id': id, 'router': router})
        return super(ApicL3Plugin, self).update_router(context, id, router)

    # REVISIT: Eliminate this wrapper?
    @n_utils.transaction_guard
    def delete_router(self, context, id):
        LOG.debug("APIC AIM L3 Plugin deleting router: %s", id)
        super(ApicL3Plugin, self).delete_router(context, id)

    def _include_router_extn_attr(self, session, router):
        attr = self.get_router_extn_db(session, router['id'])
        router.update(attr)

    def _include_router_extn_attr_bulk(self, session, routers):
        router_ids = [router['id'] for router in routers]
        attr_dict = self.get_router_extn_db_bulk(session, router_ids)
        for router in routers:
            router.update(attr_dict[router['id']] if router['id'] in attr_dict
                          else {l3_ext.EXTERNAL_PROVIDED_CONTRACTS: [],
                                l3_ext.EXTERNAL_CONSUMED_CONTRACTS: []})

    @n_utils.transaction_guard
    def add_router_interface(self, context, router_id, interface_info):
        LOG.debug("APIC AIM L3 Plugin adding interface %(interface)s "
                  "to router %(router)s",
                  {'interface': interface_info, 'router': router_id})
        # REVISIT: Remove override flag when no longer needed for GBP.
        context.override_network_routing_topology_validation = (
            interface_info.get(
                l3_ext.OVERRIDE_NETWORK_ROUTING_TOPOLOGY_VALIDATION))
        # REVISIT: The REST API layer unwraps these exceptions, such
        # as MechanismDriverError raised by ML2, returning the
        # inner_exceptions to the client.  But many UTs call these L3
        # API methods directly, and need the proper exception to be
        # raised. Changing the UTs to unwrap raised exceptions
        # themselves would eliminate the need to do it here.
        try:
            info = super(ApicL3Plugin, self).add_router_interface(
                context, router_id, interface_info)
        except exceptions.MultipleExceptions as e:
            inner = e.inner_exceptions
            raise inner[0] if inner else e
        finally:
            del context.override_network_routing_topology_validation
        # REVISIT: This could be moved to create_port_postcommit and
        # update_port_postcommit.
        self._core_plugin.update_port(
            context, info['port_id'],
            {'port': {portbindings.HOST_ID: md.FABRIC_HOST_ID}})
        return info

    @n_utils.transaction_guard
    def remove_router_interface(self, context, router_id, interface_info):
        LOG.debug("APIC AIM L3 Plugin removing interface %(interface)s "
                  "from router %(router)s",
                  {'interface': interface_info, 'router': router_id})
        # REVISIT: The REST API layer unwraps these exceptions, such
        # as MechanismDriverError raised by ML2, returning the
        # inner_exceptions to the client.  But many UTs call these L3
        # API methods directly, and need the proper exception to be
        # raised. Changing the UTs to unwrap raised exceptions
        # themselves would eliminate the need to do it here.
        try:
            info = super(ApicL3Plugin, self).remove_router_interface(
                context, router_id, interface_info)
        except exceptions.MultipleExceptions as e:
            inner = e.inner_exceptions
            raise inner[0] if inner else e
        return info

    @n_utils.transaction_guard
    def create_floatingip(self, context, floatingip):
        fip = floatingip['floatingip']
        # REVISIT: This ensure_tenant call probably isn't needed, as
        # FIPs don't map directly to any AIM resources. If it is
        # needed, it could me moved to the FLOATING_IP.BEFORE_CREATE
        # callback in rocky and newer.
        self._md.ensure_tenant(context, fip['tenant_id'])
        with db_api.CONTEXT_READER.using(context):
            # Verify that subnet is not a SNAT host-pool. This could
            # be done from a FLOATING_IP.PRECOMMIT_CREATE callback,
            # but that callback is made after a FIP port has been
            # allocated from the subnet. An exception would cause that
            # port to be deleted, but we are better off not trying to
            # allocate from the SNAT subnet in the first place.
            self._md.check_floatingip_external_address(context, fip)
        if fip.get('subnet_id') or fip.get('floating_ip_address'):
            result = super(ApicL3Plugin, self).create_floatingip(
                context, floatingip)
        else:
            # Iterate over non SNAT host-pool subnets and try to
            # allocate an address.
            with db_api.CONTEXT_READER.using(context):
                other_subs = self._md.get_subnets_for_fip(context, fip)
            result = None
            for ext_sn in other_subs:
                fip['subnet_id'] = ext_sn
                try:
                    result = (super(ApicL3Plugin, self).create_floatingip(
                        context, floatingip))
                    break
                except exceptions.IpAddressGenerationFailure:
                    LOG.info('No more floating IP addresses available '
                             'in subnet %s', ext_sn)
            if not result:
                raise exceptions.IpAddressGenerationFailure(
                    net_id=fip['floating_network_id'])
        # REVISIT: Replace with FLOATING_IP.AFTER_UPDATE callback,
        # which is called after creation as well, in ocata and newer
        # (called inside transaction in newton)?
        self._md.create_floatingip(context, result)
        # REVISIT: Consider using FLOATING_IP.PRECOMMIT_UPDATE
        # callback, which is called after creation as well, in queens
        # and newer, or maybe just calling update_floatingip_status
        # from the MD's create_floatingip method.
        with db_api.CONTEXT_WRITER.using(context):
            self.update_floatingip_status(
                context, result['id'], result['status'])
        return result

    @n_utils.transaction_guard
    def update_floatingip(self, context, id, floatingip):
        old_fip = self.get_floatingip(context, id)
        result = super(ApicL3Plugin, self).update_floatingip(
            context, id, floatingip)
        # REVISIT: Replace with FLOATING_IP.AFTER_UPDATE callback in
        # ocata and newer (called inside transaction in newton)?
        self._md.update_floatingip(context, old_fip, result)
        # REVISIT: Consider using FLOATING_IP.PRECOMMIT_UPDATE
        # callback in queens and newer, or maybe just calling
        # update_floatingip_status from the MD's update_floatingip
        # method.
        if old_fip['status'] != result['status']:
            with db_api.CONTEXT_WRITER.using(context):
                self.update_floatingip_status(
                    context, result['id'], result['status'])
        return result

    @n_utils.transaction_guard
    def delete_floatingip(self, context, id):
        old_fip = self.get_floatingip(context, id)
        super(ApicL3Plugin, self).delete_floatingip(context, id)
        # REVISIT: Replace with FLOATING_IP.AFTER_DELETE callback in
        # queens and newer?
        self._md.delete_floatingip(context, old_fip)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import operator
import os

import oslo_messaging as messaging
from oslo_serialization import jsonutils
import requests
import six

from gbpservice.contrib.nfp.configurator.agents import agent_base
from gbpservice.contrib.nfp.configurator.lib import constants as common_const
from gbpservice.contrib.nfp.configurator.lib import fw_constants as const
from gbpservice.contrib.nfp.configurator.lib import utils as load_driver
from gbpservice.nfp.core import event as nfp_event
from gbpservice.nfp.core import log as nfp_logging
from gbpservice.nfp.core import module as nfp_api

LOG = nfp_logging.getLogger(__name__)


class FwaasRpcSender(agent_base.AgentBaseEventHandler):
    """ Implements Fwaas response path to Neutron plugin.

    Methods of this class are invoked by the FwaasEventHandler class
    for sending response from driver to the Fwaas Neutron plugin.

    """

    def __init__(self, sc, host, drivers, rpcmgr):
        super(FwaasRpcSender, self).__init__(sc, drivers, rpcmgr)
        self.host = host

    def set_firewall_status(self, agent_info,
                            firewall_id, status, firewall=None):
        """ Enqueues the response from FwaaS operation to neutron plugin.

        :param context: Neutron context
        :param firewall_id: id of firewall resource
        :param status: ACTIVE/ ERROR

        """

        msg = {'info': {'service_type': const.SERVICE_TYPE,
                        'context': agent_info['context']},
               'notification': [{
                   'resource': agent_info['resource'],
                   'data': {'firewall_id': firewall_id,
                            'host': self.host,
                            'status': status,
                            'notification_type': (
                                'set_firewall_status')}}]
               }
        LOG.info("Sending Notification 'Set Firewall Status' to "
                 "Orchestrator for firewall: %(fw_id)s with status:"
                 "%(status)s",
                 {'fw_id': firewall_id,
                  'status': status})
        self.notify._notification(msg)

    def firewall_deleted(self, agent_info, firewall_id, firewall=None):
        """ Enqueues the response from FwaaS operation to neutron plugin.

        :param context: Neutron context
        :param firewall_id: id of firewall resource

        """

        msg = {'info': {'service_type': const.SERVICE_TYPE,
                        'context': agent_info['context']},
               'notification': [{
                   'resource': agent_info['resource'],
                   'data': {'firewall_id': firewall_id,
                            'host': self.host,
                            'notification_type': (
                                'firewall_deleted')}}]
               }
        LOG.info("Sending Notification 'Firewall Deleted' to "
                 "Orchestrator for firewall: %(fw_id)s ",
                 {'fw_id': firewall_id})
        self.notify._notification(msg)


class FWaasRpcManager(agent_base.AgentBaseRPCManager):
    """ Implements FWaasRpcManager class which receives requests
        from Configurator to Agent.

    Methods of this class are invoked by the configurator. Events are
    created according to the requests received and enqueued to worker queues.

    """

    RPC_API_VERSION = '1.0'
    target = messaging.Target(version=RPC_API_VERSION)

    def __init__(self, sc, conf):
        """Instantiates child and parent class objects.

        :param sc: Service Controller object that is used to communicate
        with process model core file.
        :param conf: Configuration object that is used for configuration
        parameter access.

        """

        super(FWaasRpcManager, self).__init__(sc, conf)

    def _create_event(self, context, firewall, host, method):
        """ Creates and enqueues the events to the worker queues.

        :param context: Neutron context
        :param firewall: Firewall resource object from neutron fwaas plugin
        :param host: Name of the host machine
        :param method: CREATE_FIREWALL/UPDATE_FIREWALL/DELETE_FIREWALL

        """

        # To solve the huge data issue with firewalls,
        # especially with 250 firewall rule test which
        # gets multipled with each consumer in the chain.
        # Even the zipped data is huge and cannot be sent
        # over pipe. Writing it to file here and event handler
        # will read it from file and process further.

        filename = "/tmp/" + firewall['id']
        with open(filename, 'w') as f:
            f.write(jsonutils.dumps(firewall))

        arg_dict = {'context': context,
                    'firewall': {'file_path': filename},
                    'host': host}
        # REVISIT(mak): How to send large data ?
        # New API required to send over unix sockert ?
        context['service_info'] = {}

        # ev = self.sc.new_event(id=method, data={}, key=None)
        ev = self.sc.new_event(id=method, data=arg_dict, key=None)
        self.sc.post_event(ev)

    def create_firewall(self, context, firewall, host):
        """ Receives request to create firewall from configurator

        """

        LOG.info("Received request 'Create Firewall'.")
        self._create_event(context, firewall,
                           host, const.FIREWALL_CREATE_EVENT)

    def update_firewall(self, context, firewall, host):
        """ Receives request to update firewall from configurator

        """
        LOG.info("Received request 'Update Firewall'.")
        self._create_event(context, firewall,
                           host, const.FIREWALL_UPDATE_EVENT)

    def delete_firewall(self, context, firewall, host):
        """ Receives request to delete firewall from configurator

        """
        LOG.info("Received request 'Delete Firewall'.")
        self._create_event(context, firewall,
                           host, const.FIREWALL_DELETE_EVENT)


class FWaasEventHandler(nfp_api.NfpEventHandler):
    """ Handler class which invokes firewall driver methods

    Worker processes dequeue the worker queues and invokes the
    appropriate handler class methods for Fwaas methods.

    """

    def __init__(self, sc, drivers, rpcmgr, conf):
        """ Instantiates class object.

        :param sc: Service Controller object that is used to communicate
        with process model core file.
        :param drivers: dictionary of driver name to object mapping
        :param rpcmgr: FwaasRpcManager class object

        """

        self.sc = sc
        self.conf = conf
        self.drivers = drivers
        self.host = self.conf.host
        self.rpcmgr = rpcmgr
        self.plugin_rpc = FwaasRpcSender(sc, self.host,
                                         self.drivers, self.rpcmgr)

    def _get_driver(self, service_vendor, service_feature):
        """ Retrieves driver object given the service type

        """

        driver_id = const.SERVICE_TYPE + service_vendor + service_feature
        return self.drivers[driver_id]

    def _is_firewall_rule_exists(self, fw):
        """ Checks if firewall rules are present in the request data

        :param fw: Firewall resource object

        """

        if not fw['firewall_rule_list']:
            return False
        else:
            return True

    def handle_event(self, ev):
        """ Demultiplexes the firewall request to appropriate
        driver methods.

        :param ev: event object sent from process model event handler

        """

        try:
            msg = ("Handling event %s" % (ev.id))
            LOG.info(msg)

            # The context here in ev.data is the neutron context that was
            # renamed to context in the agent_base. This erstwhile
            # neutron context contains the agent info which in turn contains
            # the API context alongside other relevant information like
            # service vendor and type. Agent info is constructed inside
            # the demuxer library.

            if ev.data['firewall'].get('file_path', None):
                filename = ev.data['firewall']['file_path']
                string = str()
                with open(filename, 'r') as f:
                    string = f.read()
                ev.data['firewall'] = jsonutils.loads(string)
                try:
                    os.remove(filename)
                except Exception as e:
                    msg = ("Exception while removing the file %r, "
                           "with error: %r" % (filename, e))
                    LOG.error(msg)

            agent_info = ev.data['context']['agent_info']
            service_vendor = agent_info['service_vendor']
            service_feature = agent_info.get('service_feature', '')
            driver = self._get_driver(service_vendor, service_feature)
            LOG.info("Invoking driver with service vendor:"
                     "%(service_vendor)s ",
                     {'service_vendor': service_vendor})
            self.method = getattr(driver, "%s" % (ev.id.lower()))
            self.invoke_driver_for_plugin_api(ev)
            msg = ("Handled event %s successfully" % (ev.id))
            LOG.info(msg)
        except Exception as err:
            msg = ("Failed handling event: %s. Reason %s"
                   % (ev.id, str(err).capitalize()))
            LOG.error(msg)

    def _remove_duplicate_fw_rules(self, rules):
        """ Removes duplicate rules from the rules list. """
        # 'description' filter field needs to be added if required
        filter_keys = ['action', 'destination_ip_address', 'destination_port',
                       'enabled', 'ip_version', 'protocol',
                       'source_ip_address', 'source_port', 'shared']
        filter_rules = []
        for rule in rules:
            filter_rules.append({k: rule[k] for k in filter_keys})

        unique_rules = [dict(tupleized) for tupleized in set(
            tuple(rule.items()) for rule in filter_rules)]
        result = []
        for d1 in unique_rules:
            for d2 in rules:
                if d1.viewitems() <= d2.viewitems():
                    result.append(d2)
                    break
        result.sort(key=operator.itemgetter('position'))
        for index, x in enumerate(result):
            x['position'] = index + 1
        return result

    def invoke_driver_for_plugin_api(self, ev):
        """ Invokes the appropriate driver methods

        :param ev: event object sent from process model event handler

        """

        context = ev.data['context']
        agent_info = context.get('agent_info')
        firewall = ev.data.get('firewall')
        host = ev.data.get('host')
        firewall['firewall_rule_list'] = self._remove_duplicate_fw_rules(
            firewall['firewall_rule_list'])

        if ev.id == const.FIREWALL_CREATE_EVENT:
            if not self._is_firewall_rule_exists(firewall):
                msg = ("Firewall rule list is empty, setting Firewall "
                       "status to ACTIVE %s" % (firewall))
                LOG.info(msg)
                return self.plugin_rpc.set_firewall_status(
                    agent_info, firewall['id'],
                    common_const.STATUS_ACTIVE, firewall)
            # Added to handle in service vm agents. VM agent will add
            # default DROP rule.
            # if not self._is_firewall_rule_exists(firewall):
            #     self.plugin_rpc.set_firewall_status(
            #         context, firewall['id'], const.STATUS_ACTIVE)
            try:
                status = self.method(context, firewall, host)
            except Exception as err:
                self.plugin_rpc.set_firewall_status(
                    agent_info, firewall['id'], common_const.STATUS_ERROR)
                msg = ("Failed to configure Firewall and status is "
                       "changed to ERROR. %s." % str(err).capitalize())
                LOG.error(msg)
            else:
                self.plugin_rpc.set_firewall_status(
                    agent_info, firewall['id'], status, firewall)
                msg = ("Configured Firewall and status set to %s" % status)
                LOG.info(msg)

        elif ev.id == const.FIREWALL_DELETE_EVENT:
            if not self._is_firewall_rule_exists(firewall):
                msg = ("Firewall rule list is empty, sending firewall deleted "
                       "status to plugin %s" % (firewall))
                LOG.info(msg)
                return self.plugin_rpc.firewall_deleted(
                    agent_info, firewall['id'], firewall)
            try:
                status = self.method(context, firewall, host)
            except requests.ConnectionError:
                # REVISIT(VIKASH): It can't be correct everytime
                msg = ("There is a connection error for firewall %r of "
                       "tenant %r. Assuming either there is serious "
                       "issue with VM or data path is completely "
                       "broken. For now marking that as delete."
                       % (firewall['id'], firewall['tenant_id']))
                LOG.warning(msg)
                self.plugin_rpc.firewall_deleted(
                    agent_info, firewall['id'], firewall)

            except Exception as err:
                # REVISIT(VIKASH): Is it correct to raise ? As the subsequent
                # attempt to clean will only re-raise the last one.And it
                # can go on and on and may not be ever recovered.
                self.plugin_rpc.set_firewall_status(
                    agent_info, firewall['id'], common_const.STATUS_ERROR)
                msg = ("Failed to delete Firewall and status is "
                       "changed to ERROR. %s." % str(err).capitalize())
                LOG.error(msg)
                # raise(err)
            else:
                if status == common_const.STATUS_ERROR:
                    self.plugin_rpc.set_firewall_status(
                        agent_info, firewall['id'], status)
                else:
                    msg = ("Firewall %r deleted of tenant: %r" % (
                           firewall['id'], firewall['tenant_id']))
                    LOG.info(msg)
                    self.plugin_rpc.firewall_deleted(
                        agent_info, firewall['id'], firewall)

        elif ev.id == const.FIREWALL_UPDATE_EVENT:
            if not self._is_firewall_rule_exists(firewall):
                return self.plugin_rpc.set_firewall_status(
                    agent_info,
                    common_const.STATUS_ACTIVE, firewall)
            try:
                status = self.method(context, firewall, host)
            except Exception as err:
                self.plugin_rpc.set_firewall_status(
                    agent_info, firewall['id'], common_const.STATUS_ERROR)
                msg = ("Failed to update Firewall and status is "
                       "changed to ERROR. %s." % str(err).capitalize())
                LOG.error(msg)
            else:
                self.plugin_rpc.set_firewall_status(
                    agent_info, firewall['id'], status, firewall)
                msg = ("Updated Firewall and status set to %s" % status)
                LOG.info(msg)
        else:
            msg = ("Wrong call to Fwaas event handler.")
            raise Exception(msg)


def events_init(sc, drivers, rpcmgr, conf):
    """Registers events with core service controller.

    All the events will come to handle_event method of class instance
    registered in 'handler' field.

    :param drivers: Driver instances registered with the service agent
    :param rpcmgr: Instance to receive all the RPC messages from configurator
    module.

    Returns: None

    """

    event_id_list = [const.FIREWALL_CREATE_EVENT,
                     const.FIREWALL_UPDATE_EVENT,
                     const.FIREWALL_DELETE_EVENT]
    evs = []
    for event in event_id_list:
        evs.append(nfp_event.Event(id=event, handler=FWaasEventHandler(
            sc, drivers, rpcmgr, conf)))
    sc.register_events(evs)


def load_drivers(conf):
    """Imports all the driver files corresponding to this agent.

    Returns: Dictionary of driver objects with a specified service type and
    vendor name

    """

    ld = load_driver.ConfiguratorUtils(conf)
    drivers = ld.load_drivers(const.SERVICE_TYPE)

    for service_type, driver_name in six.iteritems(drivers):
        driver_obj = driver_name(conf=conf)
        drivers[service_type] = driver_obj

    LOG.info("Firewall loaded drivers:%(drivers)s",
             {'drivers': drivers})
    return drivers


def register_service_agent(cm, sc, conf, rpcmgr):
    """Registers Fwaas service agent with configurator module.

    :param cm: Instance of configurator module
    :param sc: Instance of core service controller
    :param conf: Instance of oslo configuration
    :param rpcmgr: Instance containing RPC methods which are invoked by
    configurator module on corresponding RPC message arrival

    """

    service_type = const.SERVICE_TYPE
    cm.register_service_agent(service_type, rpcmgr)


def init_agent(cm, sc, conf):
    """Initializes Fwaas agent.

    :param cm: Instance of configuration module
    :param sc: Instance of core service controller
    :param conf: Instance of oslo configuration

    """

    try:
        drivers = load_drivers(conf)
    except Exception as err:
        msg = ("Fwaas failed to load drivers. %s"
               % (str(err).capitalize()))
        LOG.error(msg)
        raise Exception(err)
    else:
        msg = ("Fwaas loaded drivers successfully.")
        LOG.debug(msg)

    rpcmgr = FWaasRpcManager(sc, conf)
    try:
        events_init(sc, drivers, rpcmgr, conf)
    except Exception as err:
        msg = ("Fwaas Events initialization unsuccessful. %s"
               % (str(err).capitalize()))
        LOG.error(msg)
        raise Exception(err)
    else:
        msg = ("Fwaas Events initialization successful.")
        LOG.debug(msg)

    try:
        register_service_agent(cm, sc, conf, rpcmgr)
    except Exception as err:
        msg = ("Fwaas service agent registration unsuccessful. %s"
               % (str(err).capitalize()))
        LOG.error(msg)
        raise Exception(err)
    else:
        msg = ("Fwaas service agent registration successful.")
        LOG.debug(msg)

    msg = ("FIREWALL as a Service Module Initialized.")
    LOG.info(msg)


def init_agent_complete(cm, sc, conf):
    """ Initializes periodic tasks

    """

    msg = (" Firewall agent init complete")
    LOG.info(msg)
